"""Gaussian process-based minimization algorithms."""

import numpy as np

from sklearn.utils import check_random_state

from .base import base_minimize
from ..utils import cook_estimator
from ..utils import normalize_dimensions


def gp_minimize(func, dimensions, base_estimator=None,
                n_calls=100, n_random_starts=10,
                acq_func="gp_hedge", acq_optimizer="auto", x0=None, y0=None,
                random_state=None, verbose=False, callback=None,
                n_points=10000, n_restarts_optimizer=5, xi=0.01, kappa=1.96,
                noise="gaussian", n_jobs=1):
    """Bayesian optimization using Gaussian Processes.

    If every function evaluation is expensive, for instance
    when the parameters are the hyperparameters of a neural network
    and the function evaluation is the mean cross-validation score across
    ten folds, optimizing the hyperparameters by standard optimization
    routines would take for ever!

    The idea is to approximate the function using a Gaussian process.
    In other words the function values are assumed to follow a multivariate
    gaussian. The covariance of the function values are given by a
    GP kernel between the parameters. Then a smart choice to choose the
    next parameter to evaluate can be made by the acquisition function
    over the Gaussian prior which is much quicker to evaluate.

    The total number of evaluations, `n_calls`, are performed like the
    following. If `x0` is provided but not `y0`, then the elements of `x0`
    are first evaluated, followed by `n_random_starts` evaluations.
    Finally, `n_calls - len(x0) - n_random_starts` evaluations are
    made guided by the surrogate model. If `x0` and `y0` are both
    provided then `n_random_starts` evaluations are first made then
    `n_calls - n_random_starts` subsequent evaluations are made
    guided by the surrogate model.

    Parameters
    ----------
    * `func` [callable]:
        Function to minimize. Should take a single list of parameters
        and return the objective value.

        If you have a search-space where all dimensions have names,
        then you can use `ProcessOptimizer.utils.use_named_args` as a decorator
        on your objective function, in order to call it directly
        with the named arguments. See `use_named_args` for an example.

    * `dimensions` [list, shape=(n_dims,)]:
        List of search space dimensions.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

         NOTE: The upper and lower bounds are inclusive for `Integer`
         dimensions.

    * `base_estimator` [a Gaussian process estimator]:
        The Gaussian process estimator to use for optimization.
        By default, a Matern kernel is used with the following
        hyperparameters tuned.
        - All the length scales of the Matern kernel.
        - The covariance amplitude that each element is multiplied with.
        - Noise that is added to the matern kernel. The noise is assumed
          to be iid gaussian.

    * `n_calls` [int, default=100]:
        Number of calls to `func`.

    * `n_random_starts` [int, default=10]:
        Number of evaluations of `func` with random points before
        approximating it with `base_estimator`.

    * `acq_func` [string, default=`"gp_hedge"`]:
        Function to minimize over the gaussian prior. Can be either

        - `"LCB"` for lower confidence bound.
        - `"EI"` for negative expected improvement.
        - `"PI"` for negative probability of improvement.
        - `"gp_hedge"` Probabilistically choose one of the above three
          acquisition functions at every iteration. The weightage
          given to these gains can be set by `eta` through `acq_func_kwargs`.
            - The gains `g_i` are initialized to zero.
            - At every iteration,
                - Each acquisition function is optimised independently to
                  propose an candidate point `X_i`.
                - Out of all these candidate points, the next point `X_best` is
                  chosen by `softmax(eta g_i)`
                - After fitting the surrogate model with `(X_best, y_best)`,
                  the gains are updated such that `g_i -= mu(X_i)`
        - `"EIps"` for negated expected improvement per second to take into
          account the function compute time. Then, the objective function is
          assumed to return two values, the first being the objective value and
          the second being the time taken in seconds.
        - `"PIps"` for negated probability of improvement per second. The
          return type of the objective function is assumed to be similar to
          that of `"EIps

    * `acq_optimizer` [string, `"sampling"` or `"lbfgs"`, default=`"lbfgs"`]:
        Method to minimize the acquistion function. The fit model
        is updated with the optimal value obtained by optimizing `acq_func`
        with `acq_optimizer`.

        The `acq_func` is computed at `n_points` sampled randomly.

        - If set to `"auto"`, then `acq_optimizer` is configured on the
          basis of the space searched over.
          If the space is Categorical then this is set to be "sampling"`.
        - If set to `"sampling"`, then the point among these `n_points`
          where the `acq_func` is minimum is the next candidate minimum.
        - If set to `"lbfgs"`, then
              - The `n_restarts_optimizer` no. of points which the acquisition
                function is least are taken as start points.
              - `"lbfgs"` is run for 20 iterations with these points as initial
                points to find local minima.
              - The optimal of these local minima is used to update the prior.

    * `x0` [list, list of lists or `None`]:
        Initial input points.

        - If it is a list of lists, use it as a list of input points.
        - If it is a list, use it as a single initial input point.
        - If it is `None`, no initial input points are used.

    * `y0` [list, scalar or `None`]
        Evaluation of initial input points.

        - If it is a list, then it corresponds to evaluations of the function
          at each element of `x0` : the i-th element of `y0` corresponds
          to the function evaluated at the i-th element of `x0`.
        - If it is a scalar, then it corresponds to the evaluation of the
          function at `x0`.
        - If it is None and `x0` is provided, then the function is evaluated
          at each element of `x0`.

    * `random_state` [int, RandomState instance, or None (default)]:
        Set random state to something other than None for reproducible
        results.

    * `verbose` [boolean, default=False]:
        Control the verbosity. It is advised to set the verbosity to True
        for long optimization runs.

    * `callback` [callable, list of callables, optional]
        If callable then `callback(res)` is called after each call to `func`.
        If list of callables, then each callable in the list is called.

    * `n_points` [int, default=10000]:
        Number of points to sample to determine the next "best" point.
        Useless if acq_optimizer is set to `"lbfgs"`.

    * `n_restarts_optimizer` [int, default=5]:
        The number of restarts of the optimizer when `acq_optimizer`
        is `"lbfgs"`.

    * `kappa` [float, default=1.96]:
        Controls how much of the variance in the predicted values should be
        taken into account. If set to be very high, then we are favouring
        exploration over exploitation and vice versa.
        Used when the acquisition is `"LCB"`.

    * `xi` [float, default=0.01]:
        Controls how much improvement one wants over the previous best
        values. Used when the acquisition is either `"EI"` or `"PI"`.

    * `noise` [float, default="gaussian"]:
        - Use noise="gaussian" if the objective returns noisy observations.
          The noise of each observation is assumed to be iid with
          mean zero and a fixed variance.
        - If the variance is known before-hand, this can be set directly
          to the variance of the noise.
        - Set this to a value close to zero (1e-10) if the function is
          noise-free. Setting to zero might cause stability issues.

    * `n_jobs` [int, default=1]
        Number of cores to run in parallel while running the lbfgs
        optimizations over the acquisition function. Valid only
        when `acq_optimizer` is set to "lbfgs."
        Defaults to 1 core. If `n_jobs=-1`, then number of jobs is set
        to number of cores.

    Returns
    -------
    * `res` [`OptimizeResult`, scipy object]:
        The optimization result returned as a OptimizeResult object.
        Important attributes are:

        - `x` [list]: location of the minimum.
        - `fun` [float]: function value at the minimum.
        - `models`: surrogate models used for each iteration.
        - `x_iters` [list of lists]: location of function evaluation for each
           iteration.
        - `func_vals` [array]: function value for each iteration.
        - `space` [Space]: the optimization space.
        - `specs` [dict]`: the call specifications.
        - `rng` [RandomState instance]: State of the random state
           at the end of minimization.

        For more details related to the OptimizeResult object, refer
        http://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html
    """
    # Check params
    rng = check_random_state(random_state)
    space = normalize_dimensions(dimensions)

    if base_estimator is None:
        base_estimator = cook_estimator(
            "GP", space=space, random_state=rng.randint(
              0, np.iinfo(np.int32).max
              ),
            noise=noise
            )

    return base_minimize(
        func, space, base_estimator=base_estimator,
        acq_func=acq_func,
        xi=xi, kappa=kappa, acq_optimizer=acq_optimizer, n_calls=n_calls,
        n_points=n_points, n_random_starts=n_random_starts,
        n_restarts_optimizer=n_restarts_optimizer,
        x0=x0, y0=y0, random_state=rng, verbose=verbose,
        callback=callback, n_jobs=n_jobs)




from sklearn.utils import check_random_state

from .base import base_minimize
from ..utils import cook_estimator


def gbrt_minimize(func, dimensions, base_estimator=None,
                  n_calls=100, n_random_starts=10,
                  acq_func="EI", acq_optimizer="auto",
                  x0=None, y0=None, random_state=None, verbose=False,
                  callback=None, n_points=10000, xi=0.01, kappa=1.96,
                  n_jobs=1):
    """Sequential optimization using gradient boosted trees.

    Gradient boosted regression trees are used to model the (very)
    expensive to evaluate function `func`. The model is improved
    by sequentially evaluating the expensive function at the next
    best point. Thereby finding the minimum of `func` with as
    few evaluations as possible.

    The total number of evaluations, `n_calls`, are performed like the
    following. If `x0` is provided but not `y0`, then the elements of `x0`
    are first evaluated, followed by `n_random_starts` evaluations.
    Finally, `n_calls - len(x0) - n_random_starts` evaluations are
    made guided by the surrogate model. If `x0` and `y0` are both
    provided then `n_random_starts` evaluations are first made then
    `n_calls - n_random_starts` subsequent evaluations are made
    guided by the surrogate model.

    Parameters
    ----------
    * `func` [callable]:
        Function to minimize. Should take a single list of parameters
        and return the objective value.

        If you have a search-space where all dimensions have names,
        then you can use `ProcessOptimizer.utils.use_named_args` as a decorator
        on your objective function, in order to call it directly
        with the named arguments. See `use_named_args` for an example.

    * `dimensions` [list, shape=(n_dims,)]:
        List of search space dimensions.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

    * `base_estimator` [`GradientBoostingQuantileRegressor`]:
        The regressor to use as surrogate model

    * `n_calls` [int, default=100]:
        Number of calls to `func`.

    * `n_random_starts` [int, default=10]:
        Number of evaluations of `func` with random points before
        approximating it with `base_estimator`.

    * `acq_func` [string, default=`"LCB"`]:
        Function to minimize over the forest posterior. Can be either

        - `"LCB"` for lower confidence bound.
        - `"EI"` for negative expected improvement.
        - `"PI"` for negative probability of improvement.
        - ``"EIps"`` for negated expected improvement per second to take into
          account the function compute time. Then, the objective function is
          assumed to return two values, the first being the objective value and
          the second being the time taken.
        - `"PIps"` for negated probability of improvement per second.

    * `x0` [list, list of lists or `None`]:
        Initial input points.

        - If it is a list of lists, use it as a list of input points.
        - If it is a list, use it as a single initial input point.
        - If it is `None`, no initial input points are used.

    * `y0` [list, scalar or `None`]:
        Evaluation of initial input points.

        - If it is a list, then it corresponds to evaluations of the function
          at each element of `x0` : the i-th element of `y0` corresponds
          to the function evaluated at the i-th element of `x0`.
        - If it is a scalar, then it corresponds to the evaluation of the
          function at `x0`.
        - If it is None and `x0` is provided, then the function is evaluated
          at each element of `x0`.

    * `random_state` [int, RandomState instance, or None (default)]:
        Set random state to something other than None for reproducible
        results.

    * `verbose` [boolean, default=False]:
        Control the verbosity. It is advised to set the verbosity to True
        for long optimization runs.

    * `callback` [callable, optional]
        If provided, then `callback(res)` is called after call to func.

    * `n_points` [int, default=10000]:
        Number of points to sample when minimizing the acquisition function.

    * `xi` [float, default=0.01]:
        Controls how much improvement one wants over the previous best
        values. Used when the acquisition is either `"EI"` or `"PI"`.

    * `kappa` [float, default=1.96]:
        Controls how much of the variance in the predicted values should be
        taken into account. If set to be very high, then we are favouring
        exploration over exploitation and vice versa.
        Used when the acquisition is `"LCB"`.

    * `n_jobs` [int, default=1]:
        The number of jobs to run in parallel for `fit` and `predict`.
        If -1, then the number of jobs is set to the number of cores.

    Returns
    -------
    * `res` [`OptimizeResult`, scipy object]:
        The optimization result returned as a OptimizeResult object.
        Important attributes are:

        - `x` [list]: location of the minimum.
        - `fun` [float]: function value at the minimum.
        - `models`: surrogate models used for each iteration.
        - `x_iters` [list of lists]: location of function evaluation for each
           iteration.
        - `func_vals` [array]: function value for each iteration.
        - `space` [Space]: the optimization space.
        - `specs` [dict]`: the call specifications.
        - `rng` [RandomState instance]: State of the random state
           at the end of minimization.

        For more details related to the OptimizeResult object, refer
        http://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html
    """
    # Check params
    rng = check_random_state(random_state)

    if base_estimator is None:
        base_estimator = cook_estimator("GBRT", random_state=rng,
                                        n_jobs=n_jobs)
    return base_minimize(func, dimensions, base_estimator,
                         n_calls=n_calls, n_points=n_points,
                         n_random_starts=n_random_starts,
                         x0=x0, y0=y0, random_state=random_state, xi=xi,
                         kappa=kappa, acq_func=acq_func, verbose=verbose,
                         callback=callback, acq_optimizer="sampling")



from sklearn.utils import check_random_state
from .space import Real, Integer, Categorical, Space
import numpy as np


class Constraints:
    def __init__(self, constraints_list, space):
        """Constraints used when sampling for the aqcuisiiton function

        Parameters
        ----------
        * `constraints_list` [list of constraints]:
            A list of constraint objects.

        * `space` [Space]:
            A Space object.
        """

        if not isinstance(space, Space):
            space = Space(space)
        # Check that all constraints are valid:
        check_constraints(space, constraints_list)
        self.space = space
        # Lists that keep track of which dimensions has which constraints:
        self.single = [None for _ in range(space.n_dims)]
        self.inclusive = [[] for _ in range(space.n_dims)]
        self.exclusive = [[] for _ in range(space.n_dims)]
        self.sum = []
        self.conditional = []
        # A copy of the list of constraints:
        self.constraints_list = constraints_list
        # Append constraints to the lists
        for constraint in constraints_list:
            if isinstance(constraint, Single):
                self.single[constraint.dimension] = constraint
            elif isinstance(constraint, Inclusive):
                self.inclusive[constraint.dimension].append(constraint)
            elif isinstance(constraint, Exclusive):
                self.exclusive[constraint.dimension].append(constraint)
            elif isinstance(constraint, Sum):
                self.sum.append(constraint)
            elif isinstance(constraint, Conditional):
                self.conditional.append(constraint)
            else:
                raise TypeError(
                    'Constraint type {} is not recognized'.format(
                        type(constraint)))

    def __eq__(self, other):
        if isinstance(other, Constraints):
            return self.constraints_list == other.constraints_list
        else:
            return False

    def rvs(self, n_samples=1, random_state=None):
        """Draw random samples that all are valid with regards to the constraints.

        The samples are in the original space. They need to be transformed
        before being passed to a model or minimizer by `space.transform()`.

        Parameters
        ----------
        * `n_samples` [int, default=1]:
            Number of samples to be drawn from the space.

        * `random_state` [int, RandomState instance, or None (default)]:
            Set random state to something other than None for reproducible
            results.

        Returns
        -------
        * `points`: [list of lists, shape=(n_points, n_dims)]
           Points sampled from the space.
        """

        rng = check_random_state(random_state)
        # A list of the samples drawn:
        rows = []
        # Number of candidates for samples that have been checked:
        n_samples_candidates = 0

        # We keep sampling until all samples a valid with regard to the
        # constraints:
        while len(rows) < n_samples:
            columns = []
            # Iterate through all dimensions:
            for i in range(self.space.n_dims):
                dim = self.space.dimensions[i]
                # If a dimension has a "Single"-type constraint we just sample
                # that value:
                if self.single[i]:
                    column = np.full(n_samples, self.single[i].value)
                else:  # Using the default rvs() for the given dimension
                    try:
                        column = (dim.rvs(n_samples=n_samples, 
                                          random_state=rng))
                    except Exception as error:
                        print(f'''Caught an error while making random points in
                         constrained space, error is: {error}''')
                columns.append(column)

            # Transpose
            for i in range(n_samples):
                r = []
                for j in range(self.space.n_dims):
                    r.append(columns[j][i])
                # For each individual sample we check that it is valid:
                if self.validate_sample(r):
                    # We only append if it is valid
                    rows.append(r)

            n_samples_candidates += n_samples
            if n_samples_candidates > 100000 and len(rows) < 100:
                # If we have less than a 1/10.000 succes rate on the sampling
                # we throw an error
                raise RuntimeError(
                    '''Could not find valid samples in constrained space.
                    Please check that the constraints allows for valid samples
                    to be drawn''')

        # We draw more samples when needed so we only return n_samples of the
        # samples
        return rows[:n_samples]

    def validate_sample(self, sample):
        """ Validates a sample of parameter values in regards to the
        constraints.

        Parameters
        ----------
        * `sample` [list]:
            A list of values for each dimension.

        Returns
        -------
        * `is_valid`: [bool]
           Returns True for valid samples and False for non-valid samples
        """

        # We iterate through all the dimensions and check the the type of
        # constriants that are applied to a single dimensions, i.e Single,
        # Exclusive and Inclusive
        #
        # We iterate through all samples which corresponds to number of
        # dimensions:
        for dim in range(len(sample)):
            # Single constraints.
            if self.single[dim]:
                if not self.single[dim].validate_constraint(sample[dim]):
                    return False

            # Inclusive constraints.
            # Check if there is a least one inclusive constraint:
            if self.inclusive[dim]:
                # We go through all inlcusive constraints for this dimension
                # and if the value is not found to be included in any of the
                # bounds of the inclusive constraints we return false.
                value_is_valid = False
                for constraint in self.inclusive[dim]:
                    if constraint.validate_constraint(sample[dim]):
                        value_is_valid = True
                        break
                if not value_is_valid:
                    return False

            # Exclusive constraints
            for constraint in self.exclusive[dim]:
                # The first time a value is inside of the excluded bounds of
                # the exclusive constraint we return false.
                if not constraint.validate_constraint(sample[dim]):
                    return False

        # We iterate through sum constriants
        for constraint in self.sum:
            if not constraint.validate_sample(sample):
                return False
        # We iterate through Conditional constraints
        for constraint in self.conditional:
            if not constraint.validate_sample(sample):
                return False
        # If we we did not find any violaiton of the constraints we return
        # True.
        return True

    def __repr__(self):
        return "Constraints({})".format(self.constraints_list)


class Single:
    def __init__(self, dimension, value, dimension_type):
        """Constraint class of type Single.

        This constraint enforces that all values drawn for the specified
        dimension equals `value`

        Parameters
        ----------
        * `dimension` [int]:
            The index of the dimension for which constraint should be applied.

        * `value` [int, float or str]:
            The enforced value for the constraint.

        * `dimension_type` [str]:
            The type of dimension. Can be 'integer','real' or 'categorical.
            Should be the same type as the dimension it is applied to.
        """

        # Validating input parameters
        if dimension_type == 'categorical':
            if not (
                    type(value) == int
                    or
                    type(value) == str
                    or
                    type(value) == float):
                raise TypeError('Single categorical constraint must be of type int, float or str. Got {}'.format(type(value)))
        elif dimension_type == 'integer':
            if not type(value) == int:
                raise TypeError('Single integer constraint must be of type int. Got {}'.format(type(value)))
        elif dimension_type == 'real':
            if not type(value) == float:
                raise TypeError('Single real constraint must be of type float. Got {}'.format(type(value)))
        else:
            raise ValueError('`dimension_type` must be a string containing "categorical", "integer" or "real". got {}'.format(dimension_type))
        if not (type(dimension) == int):
            raise TypeError('Constraint dimension must be of type int. Got type {}'.format(type(dimension)))
        if not dimension >= 0:
            raise ValueError('Dimension can not be a negative number')

        self.value = value
        self.dimension = dimension
        self.dimension_type = dimension_type
        self.validate_constraint = self._validate_constraint
        self.validate_sample = self._validate_sample

    def _validate_sample(self, sample):
        # Returns True if sample does not violate the constraints.
        return self.validate_constraint(sample[self.dimension])

    def _validate_constraint(self, value):
        # Compares value with constraint.
        if value == self.value:
            return True
        else:
            return False

    def __repr__(self):
        return "Single(dimension={}, value={}, dimension_type={})".format(self.dimension, self.value, self.dimension_type)

    def __eq__(self, other):
        if isinstance(other, Single):
            return self.value == other.value and self.dimension == other.dimension and self.dimension_type == other.dimension_type
        else:
            return False


class Bound_constraint():
    def __init__(self, dimension, bounds, dimension_type):
        """Base class for Exclusive and Inclusive constraints"""

        # Validating input parameters
        if not type(bounds) == tuple and not type(bounds) == list:
            raise TypeError('Bounds should be a tuple or a list. Got {}'.format(type(bounds)))
        if not len(bounds) > 1:
            raise ValueError('Bounds should be a tuple or list of length > 1.')
        if not dimension_type == 'categorical':
            if not len(bounds) == 2:
                raise ValueError('Length of bounds must be 2 for non-categorical constraints. Got {}'.format(len(bounds)))
        for bound in bounds:
            if dimension_type == 'integer':
                if not type(bound) == int:
                    raise TypeError('Bounds must be of type int for integer dimension. Got {}'.format(type(bound)))
            elif dimension_type == 'real':
                if not type(bound) == float:
                    raise TypeError('Bounds must be of type float for real dimension. Got {}'.format(type(bound)))
        if not (dimension_type == 'categorical' or dimension_type == 'integer' or dimension_type == 'real'):
            raise ValueError('`dimension_type` must be a string containing "categorical", "integer" or "real". Got {}'.format(dimension_type))
        if not (type(dimension) == int):
            raise TypeError('Constraint dimension must be of type int. Got type {}'.format(type(dimension)))
        if not dimension >= 0:
            raise ValueError('Dimension can not be a negative number')

        self.bounds = tuple(bounds)  # Convert bounds to a tuple
        self.dimension = dimension
        self.dimension_type = dimension_type


class Inclusive(Bound_constraint):
    def __init__(self, dimension, bounds, dimension_type):
        super().__init__(dimension, bounds, dimension_type)
        """Constraint class of type Inclusive.

        This constraint enforces that all values drawn for the specified
        dimension are inside the bounds, with the bound values included. In
        case of categorical dimensions the constraint enforces that only values
        specified in bounds can be drawn.

        Parameters
        ----------
        * `dimension` [int]:
            The index of the dimension for which constraint should be applied.

        * `bounds` [tuple]:
            The bounds for the constraint.
            For 'real' dimensions the tuple must be of length 2 and only
            consist of integers.

            For 'integer' dimensions the tuple must be of length 2 and only
            consist of floats.

            For 'categorical' dimensions the tuple must be of length < number
            of dimensions and lenght > 1.

            The tuple can contain any combination of str, int or float

        * `dimension_type` [str]:
            The type of dimension. Can be 'integer','real' or 'categorical.
            Should be the same type as the dimension it is applied to.
        """

        # We use another validation strategy for categorical dimensions.
        if dimension_type == 'categorical':
            self.validate_constraint = self._validate_constraint_categorical
        else:
            self.validate_constraint = self._validate_constraint_real
        self.validate_sample = self._validate_sample

    def _validate_sample(self, sample):
        # Checks if a sample violates the constraints.
        # Returns True if it does not.
        return self.validate_constraint(sample[self.dimension])

    def _validate_constraint_categorical(self, value):
        if value in self.bounds:
            return True
        else:
            return False

    def _validate_constraint_real(self, value):
        if value >= self.bounds[0] and value <= self.bounds[1]:
            return True
        else:
            return False

    def __repr__(self):
        return "Inclusive(dimension={}, bounds={}, dimension_type={})".format(self.dimension, self.bounds, self.dimension_type)

    def __eq__(self, other):
        if isinstance(other, Inclusive):
            return all([a == b for a, b in zip(self.bounds, other.bounds)]) and self.dimension == other.dimension and self.dimension_type == other.dimension_type
        else:
            return False


class Exclusive(Bound_constraint):
    def __init__(self, dimension, bounds, dimension_type):
        super().__init__(dimension, bounds, dimension_type)
        """Constraint class of type Inclusive.

        This constraint enforces that all values drawn for the specified
        dimension are outside the bounds, with the bound values excluded. In
        case of categorical dimensions the constraint enforces that values
        specified in bounds can not be drawn.

        Parameters
        ----------
        * `dimension` [int]:
            The index of the dimension for which constraint should be applied.

        * `bounds` [tuple]:
            The bounds for the constraint.
            For 'real' dimensions the tuple must be of length 2 and only
            consist of integers.

            For 'integer' dimensions the tuple must be of length 2 and only
            consist of floats.

            For 'categorical' dimensions the tuple must be of length < number
            of dimensions and lenght > 1.

            The tuple can contain any combination of str, int or float

        * `dimension_type` [str]:
            The type of dimension. Can be 'integer','real' or 'categorical.
            Should be the same type as the dimension it is applied to.
        """

        # We use another validation strategy for categorical dimensions.
        if dimension_type == 'categorical':
            self.validate_constraint = self._validate_constraint_categorical
        else:
            self.validate_constraint = self._validate_constraint_real
        self.validate_sample = self._validate_sample

    def _validate_sample(self, sample):
        # Returns True if sample does not violate the constraints.
        return self.validate_constraint(sample[self.dimension])

    def _validate_constraint_categorical(self, value):
        if value in self.bounds:
            return False
        else:
            return True

    def _validate_constraint_real(self, value):
        if value >= self.bounds[0] and value <= self.bounds[1]:
            return False
        else:
            return True

    def __repr__(self):
        return "Exclusive(dimension={}, bounds={}, dimension_type={})".format(self.dimension, self.bounds, self.dimension_type)

    def __eq__(self, other):
        if isinstance(other, Exclusive):
            return all([a == b for a, b in zip(self.bounds, other.bounds)]) and self.dimension == other.dimension and self.dimension_type == other.dimension_type
        else:
            return False


class Sum():
    def __init__(self, dimensions, value, less_than=True):
        """Constraint class of type Sum.

        This constraint enforces that the sum of all values drawn for the
        specified dimensions, is less_than or equal to a value. Can only be
        used with integer or real dimensions.

        Parameters
        ----------
        * `dimensions` [list of ints]:
            A list of integers coresponding to the index of the dimensions that
            should be summed

        * `value` [float or int]:
            The value for which the sum should be less than or equal to.

        * `less_than` [bool, default=True]:
            If true then the sum should be less than or equal to the value. If
            False the sum should be greater than or equal to the value.
        """

        if not (type(dimensions) == tuple or type(dimensions) == list):
            raise TypeError('Argument `dimensions` must be of type tuple or list. Got {}'.format(type(dimensions)))
        if not len(dimensions) > 1:
            raise ValueError('Argument `dimensions` must have a lenght of more than 1. Got {}'.format(len(dimensions)))
        if not all(type(dim) == int for dim in dimensions):
            raise TypeError('dimension indices must all be of type int.')
        if not all(dim >= 0 for dim in dimensions):
            raise ValueError('Dimension index must be positive')
        if not (type(value) == int or type(value) == float):
            raise TypeError('Argument `value` must be of type float or int. Got {}'.format(type(value)))
        if not type(less_than) == bool:
            raise TypeError('Argument `less_than` must be of type bool (True or False). Got {}'.format(type(less_than)))

        self.dimensions = tuple(dimensions)
        self.value = value
        self.less_than = less_than

        self.validate_sample = self._validate_sample

    def _validate_sample(self, sample):
        # Returns True if sample does not violate the constraints.
        if self.less_than:
            return np.sum([sample[dim] for dim in self.dimensions]) <= self.value
        else:
            return np.sum([sample[dim] for dim in self.dimensions]) >= self.value

    def __repr__(self):
        return "Sum(dimensions={}, value={}, less_than={})".format(self.dimensions, self.value, self.less_than)

    def __eq__(self, other):
        if isinstance(other, Sum):
            return all([a == b for a, b in zip(self.dimensions, other.dimensions)]) and self.value == other.value and self.less_than == other.less_than
        else:
            return False


class Conditional():
    def __init__(self, condition, if_true=None, if_false=None):
        ''' Constraint class of type Conditional

        This constraint enforces other constraints depending on the
        'condition'. If condition is satisfied the if_true constraint is
        applied. If not then the if_false constraint is applied.

        Parameters
        ----------
        * `condition` [Constraint]:
            A constraint that defines the condition.
            Condition can not be a Conditional constraint.

        * `if_true` [Constraint or tuple/list of constraints, deault=None]:
            If the constraint from the condition-argument is satisfied then
            this constraint is applied.

        * `if_false` [Constraint or tuple/list of constraints, deault=None]:
            If the constraint from the condition-argument is not satisfied then
            this constraint is applied.
        '''

        if isinstance(condition, Conditional):
            raise TypeError('Condition can not be a conditional constraint')
        check_is_constraint(condition)
        if if_true:
            if type(if_true) == tuple or type(if_true) == list:
                # Check all constraints
                for constraint in if_true:
                    check_is_constraint(constraint)
                if_true = tuple(if_true)
            else:
                check_is_constraint(if_true)
                # Convert to tuple if only one constraint has been passed
                if_true = tuple([if_true])
        else:
            # Convert "None" to empty tuple
            if_true = ()
        if if_false:
            if type(if_false) == tuple or type(if_false) == list:
                for constraint in if_false:
                    check_is_constraint(constraint)
                    if_false = tuple(if_false)
            else:
                check_is_constraint(if_false)
                if_false = tuple([if_false])
        else:
            if_false = ()

        self.condition = condition
        self.if_true = if_true
        self.if_false = if_false
        self.validate_sample = self._validate_sample

    def _validate_sample(self, sample):
        # Returns True if sample does not violate the constraints.
        if self.condition.validate_sample(sample):
            # Condition evaluates to true
            if self.if_true:
                # Iterate through all constrains in if_true:
                for constraint in self.if_true:
                    if not constraint.validate_sample(sample):
                        # The first time a constraint fails we return False
                        return False
                # If no constraints failed we return True
                return True
            else:  # If there is no constraints we return True
                return True
        else:
            if self.if_false:
                for constraint in self.if_false:
                    if not constraint.validate_sample(sample):
                        return False
                return True
            else:
                return True

    def __repr__(self):
        return "Conditional(condition={}, if_true={}, if_false={})".format(self.condition, self.if_true, self.if_false)

    def __eq__(self, other):
        if isinstance(other, Conditional):
            return self.condition == other.condition and self.if_true == other.if_true and self.if_false == other.if_false
        else:
            return False


def check_constraints(space, constraints):
    """ Checks if list of constraints is valid when compared with the
    dimensions of space. Throws errors otherwise.

        Parameters
        ----------
        * `space` [Space]:
            A Space object.

        * `constraints` [list]:
            A list of constraint objects.
    """

    if not type(constraints) == list:
        raise TypeError('Constraints must be a list of constraints. Got {}'.format(type(constraints)))
    n_dims = space.n_dims

    single_constraints = [False] * n_dims
    for constraint in constraints:

        # Check if constraints are inside bounds of space and if more than one
        # single constraint has been applied to the same dimension
        if isinstance(constraint, Single):
            check_dim_and_space(space, constraint)
            ind_dim = constraint.dimension
            if single_constraints[ind_dim] is True:
                raise IndexError('Can not add more than one Singe-type constraint to dimension {}'.format(ind_dim))
            single_constraints[ind_dim] = True
            space_dim = space.dimensions[constraint.dimension]
            check_value(space_dim, constraint.value)
        elif isinstance(constraint, Inclusive) or isinstance(constraint, Exclusive):
            check_dim_and_space(space, constraint)
            space_dim = space.dimensions[constraint.dimension]
            check_bounds(space_dim, constraint.bounds)
        elif isinstance(constraint, Sum):
            if not all(dim < n_dims for dim in constraint.dimensions):
                raise IndexError('Dimension index exceeds number of dimensions')
            for ind_dim in constraint.dimensions:
                if isinstance(space.dimensions[ind_dim], Categorical):
                    raise ValueError('Sum constraint can not be applid to categorical dimension: {}'.format(space.dimensions[ind_dim]))
        elif isinstance(constraint, Conditional):
            # We run check_constraints on each constraint instance in the
            # conditional constraint. We only check them if they are not None
            if constraint.condition:
                check_constraints(space, [constraint.condition])
            if constraint.if_true:
                # Check all constraints in the if_true tuple
                for constraint_if_true in constraint.if_true:
                    check_constraints(space, [constraint_if_true])
            if constraint.if_false:
                for constraint_if_false in constraint.if_false:
                    check_constraints(space, [constraint_if_false])
        else:
            raise TypeError('Constraints must be of type "Single", "Exlusive", "Inclusive", "Sum" or "Conditional". Got {}'.format(type(constraint)))


def check_dim_and_space(space, constraint):
    n_dims = space.n_dims
    ind_dim = constraint.dimension  # Index of the dimension
    if not ind_dim < n_dims:
        raise IndexError('Dimension index {} out of range for n_dims = {}'.format(ind_dim, n_dims))
    space_dim = space.dimensions[constraint.dimension]

    # Check if space dimensions types are the same as constraint dimension types
    if isinstance(space_dim, Real):
        if not constraint.dimension_type == 'real':
            raise TypeError('Constraint for real dimension {} must be of dimension_type real. Got {}'.format(ind_dim, constraint.dimension_type))
    elif isinstance(space_dim, Integer):
        if not constraint.dimension_type == 'integer':
            raise TypeError('Constraint for integer dimension {} must be of dimension_type integer. Got {}'.format(ind_dim, constraint.dimension_type))
    elif isinstance(space_dim, Categorical):
        if not constraint.dimension_type == 'categorical':
            raise TypeError('Constraint for categorical dimension {} must be of dimension_type categorical. Got {}'.format(ind_dim, constraint.dimension_type))
    else:
        raise TypeError('Can not find valid dimension for' + str(space_dim))


def check_bounds(dim, bounds):
    """ Checks if bounds are included in the dimension.
        Throws errors otherwise.

        Parameters
        ----------
        * `dim` [Dimension]:
            A Space object.

        * `bounds` [tuple]:
            A tuple of int, float or str depending on the type of dimension.
        """

    # Real or integer dimensions.
    if isinstance(dim, Real) or isinstance(dim, Integer):
        for value in bounds:
            if value < dim.low or value > dim.high:
                raise ValueError('Bounds {} exceeds bounds of space {}'.format(bounds, [dim.low, dim.high]))
    else:  # Categorical dimensions.
        for value in bounds:
            if value not in dim.categories:
                raise ValueError('Categorical value {} is not in space with categories {}'.format(value, dim.categories))


def check_value(dim, value):
    """ Checks if value are included in the bounds of the dimension.
        Throws errors otherwise.

        Parameters
        ----------
        * `dim` [Dimension]:
            A Space object.

        * `value` [int, float or str]:
            The type of `value` depends on the type of dimension.
        """

    # Real or integer dimension
    if isinstance(dim, Real) or isinstance(dim, Integer):
        if value < dim.low or value > dim.high:
            raise ValueError('Value {} exceeds bounds of space {}'.format(value, [dim.low, dim.high]))
    else:  # Categorical dimension.
        if value not in dim.categories:
            raise ValueError('Categorical value {} is not in space with categoreis {}'.format(value, dim.categories))


def check_is_constraint(constraint):
    ''' Checks if constraint is a valid constraint type. Throws error otherwise.
    Types can be Single, Inclusive, Exclusive, Sum and Conditional'''

    if not (isinstance(constraint, Single) or isinstance(constraint, Inclusive)
            or isinstance(constraint, Exclusive) or isinstance(constraint, Sum)
            or isinstance(constraint, Conditional)):
        raise TypeError('Constraint must be of type Inclusive, Exlusive, Single, Sum or Conditional. Got {}'.format(type(constraint)))



# -*- coding: utf-8 -*-

# The following has been copied from MOBOpt
# (https://github.com/ppgaluzio/MOBOpt)

import numpy as np

import array
import random

from deap import base
from deap import creator
from deap import tools

FirstCall = True


def uniform(bounds):
    return [random.uniform(b[0], b[1]) for b in bounds]


def NSGAII(NObj, objective, pbounds, seed=None, NGEN=50, MU=100, CXPB=0.9):
    random.seed(seed)

    global FirstCall
    if FirstCall:
        creator.create("FitnessMin", base.Fitness, weights=(-1.0,)*NObj)
        creator.create("Individual", array.array, typecode='d',
                       fitness=creator.FitnessMin)
        FirstCall = False
    toolbox = base.Toolbox()

    NDIM = len(pbounds)

    toolbox.register("attr_float", uniform, pbounds)

    toolbox.register("individual",
                     tools.initIterate,
                     creator.Individual,
                     toolbox.attr_float)

    toolbox.register("population", tools.initRepeat, list, toolbox.individual)

    toolbox.register("evaluate", objective)

    toolbox.register("mate",
                     tools.cxSimulatedBinaryBounded,
                     low=pbounds[:, 0].tolist(),
                     up=pbounds[:, 1].tolist(),
                     eta=20.0)

    toolbox.register("mutate",
                     tools.mutPolynomialBounded,
                     low=pbounds[:, 0].tolist(),
                     up=pbounds[:, 1].tolist(),
                     eta=20.0,
                     indpb=1.0/NDIM)

    toolbox.register("select", tools.selNSGA2)

    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("min", np.min, axis=0)
    stats.register("max", np.max, axis=0)

    logbook = tools.Logbook()
    logbook.header = "gen", "evals", "std", "min", "avg", "max"

    pop = toolbox.population(n=MU)

    # Evaluate the individuals with an invalid fitness
    invalid_ind = [ind for ind in pop if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit

    # This is just to assign the crowding distance to the individuals
    # no actual selection is done
    pop = toolbox.select(pop, len(pop))

    record = stats.compile(pop)
    logbook.record(gen=0, evals=len(invalid_ind), **record)

    # Begin the generational process
    for gen in range(1, NGEN):
        # Vary the population
        offspring = tools.selTournamentDCD(pop, len(pop))
        offspring = [toolbox.clone(ind) for ind in offspring]

        for ind1, ind2 in zip(offspring[::2], offspring[1::2]):
            if random.random() <= CXPB:
                toolbox.mate(ind1, ind2)

            toolbox.mutate(ind1)
            toolbox.mutate(ind2)
            del ind1.fitness.values, ind2.fitness.values

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # Select the next generation population
        pop = toolbox.select(pop + offspring, MU)
        record = stats.compile(pop)
        logbook.record(gen=gen, evals=len(invalid_ind), **record)
        # print(logbook.stream)

    front = np.array([ind.fitness.values for ind in pop])

    return pop, logbook, front



"""
`ProcessOptimizer`, is a simple and efficient library to minimize (very)
expensive and noisy black-box functions. It implements several methods for
sequential model-based optimization. `ProcessOptimizer` is reusable in many
contexts and accessible.

## Install

pip install ProcessOptimizer

## Development

The library is still experimental and under heavy development.

"""

from . import acquisition
from .model_systems import benchmarks
from . import callbacks
from . import learning
from . import optimizer

from . import space
from .optimizer import dummy_minimize
from .optimizer import forest_minimize
from .optimizer import gbrt_minimize
from .optimizer import gp_minimize
from .optimizer import Optimizer
from .searchcv import BayesSearchCV
from .space import Space
from .utils import dump
from .utils import expected_minimum
from .utils import expected_minimum_random_sampling
from .utils import load
from .utils import cook_estimator, create_result, y_coverage
from .plots import plot_objective, plot_objectives
from .plots import plot_evaluations, plot_convergence
from .plots import plot_Pareto, plot_expected_minimum_convergence

__version__ = "0.7.8"


__all__ = (
    "acquisition",
    "benchmarks",
    "callbacks",
    "learning",
    "optimizer",
    "plots",
    "space",
    "gp_minimize",
    "dummy_minimize",
    "forest_minimize",
    "gbrt_minimize",
    "Optimizer",
    "dump",
    "load",
    "cook_estimator",
    "create_result",
    "expected_minimum",
    "expected_minimum_random_sampling",
    "BayesSearchCV",
    "Space"
    "Space",
    "plot_objective",
    "plot_objectives",
    "plot_evaluations",
    "plot_convergence",
    "plot_Pareto",
    "y_coverage",
    "plot_expected_minimum_convergence"
)



import numpy as np
import warnings

from scipy.stats import norm


def gaussian_acquisition_1D(X, model, y_opt=None, acq_func="LCB",
                            acq_func_kwargs=None, return_grad=True):
    """
    A wrapper around the acquisition function that is called by fmin_l_bfgs_b.

    This is because lbfgs allows only 1-D input.
    """
    return _gaussian_acquisition(np.expand_dims(X, axis=0),
                                 model, y_opt, acq_func=acq_func,
                                 acq_func_kwargs=acq_func_kwargs,
                                 return_grad=return_grad)


def _gaussian_acquisition(X, model, y_opt=None, acq_func="LCB",
                          return_grad=False, acq_func_kwargs=None):
    """
    Wrapper so that the output of this function can be
    directly passed to a minimizer.
    """
    # Check inputs
    X = np.asarray(X)
    if X.ndim != 2:
        raise ValueError("X is {}-dimensional, however,"
                         " it must be 2-dimensional.".format(X.ndim))

    if acq_func_kwargs is None:
        acq_func_kwargs = dict()
    xi = acq_func_kwargs.get("xi", 0.01)
    kappa = acq_func_kwargs.get("kappa", 1.96)

    # Evaluate acquisition function
    per_second = acq_func.endswith("ps")
    if per_second:
        model, time_model = model.estimators_

    if acq_func == "LCB":
        func_and_grad = gaussian_lcb(X, model, kappa, return_grad)
        if return_grad:
            acq_vals, acq_grad = func_and_grad
        else:
            acq_vals = func_and_grad

    elif acq_func in ["EI", "PI", "EIps", "PIps"]:
        if acq_func in ["EI", "EIps"]:
            func_and_grad = gaussian_ei(X, model, y_opt, xi, return_grad)
        else:
            func_and_grad = gaussian_pi(X, model, y_opt, xi, return_grad)

        if return_grad:
            acq_vals = -func_and_grad[0]
            acq_grad = -func_and_grad[1]
        else:
            acq_vals = -func_and_grad

        if acq_func in ["EIps", "PIps"]:

            if return_grad:
                mu, std, mu_grad, std_grad = time_model.predict(
                    X, return_std=True, return_mean_grad=True,
                    return_std_grad=True)
            else:
                mu, std = time_model.predict(X, return_std=True)

            # acq = acq / E(t)
            inv_t = np.exp(-mu + 0.5*std**2)
            acq_vals *= inv_t

            # grad = d(acq_func) * inv_t + (acq_vals *d(inv_t))
            # inv_t = exp(g)
            # d(inv_t) = inv_t * grad(g)
            # d(inv_t) = inv_t * (-mu_grad + std * std_grad)
            if return_grad:
                acq_grad *= inv_t
                acq_grad += acq_vals * (-mu_grad + std*std_grad)

    else:
        raise ValueError("Acquisition function not implemented.")

    if return_grad:
        return acq_vals, acq_grad
    return acq_vals


def gaussian_lcb(X, model, kappa=1.96, return_grad=False):
    """
    Use the lower confidence bound to estimate the acquisition
    values.

    The trade-off between exploitation and exploration is left to
    be controlled by the user through the parameter ``kappa``.

    Parameters
    ----------
    * `X` [array-like, shape=(n_samples, n_features)]:
        Values where the acquisition function should be computed.

    * `model` [sklearn estimator that implements predict with ``return_std``]:
        The fit estimator that approximates the function through the
        method ``predict``.
        It should have a ``return_std`` parameter that returns the standard
        deviation.

    * `kappa`: [float, default 1.96 or 'inf']:
        Controls how much of the variance in the predicted values should be
        taken into account. If set to be very high, then we are favouring
        exploration over exploitation and vice versa.
        If set to 'inf', the acquisition function will only use the variance
        which is useful in a pure exploration setting.
        Useless if ``method`` is set to "LCB".

    * `return_grad`: [boolean, optional]:
        Whether or not to return the grad. Implemented only for the case where
        ``X`` is a single sample.

    Returns
    -------
    * `values`: [array-like, shape=(X.shape[0],)]:
        Acquisition function values computed at X.

    * `grad`: [array-like, shape=(n_samples, n_features)]:
        Gradient at X.
    """
    # Compute posterior.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        if return_grad:
            mu, std, mu_grad, std_grad = model.predict(
                X, return_std=True, return_mean_grad=True,
                return_std_grad=True)

            if kappa == "inf":
                return -std, -std_grad
            return mu - kappa * std, mu_grad - kappa * std_grad

        else:
            mu, std = model.predict(X, return_std=True)
            if kappa == "inf":
                return -std
            return mu - kappa * std


def gaussian_pi(X, model, y_opt=0.0, xi=0.01, return_grad=False):
    """
    Use the probability of improvement to calculate the acquisition values.

    The conditional probability `P(y=f(x) | x)`form a gaussian with a
    certain mean and standard deviation approximated by the model.

    The PI condition is derived by computing ``E[u(f(x))]``
    where ``u(f(x)) = 1``, if ``f(x) < y_opt`` and ``u(f(x)) = 0``,
    if``f(x) > y_opt``.

    This means that the PI condition does not care about how "better" the
    predictions are than the previous values, since it gives an equal reward
    to all of them.

    Note that the value returned by this function should be maximized to
    obtain the ``X`` with maximum improvement.

    Parameters
    ----------
    * `X` [array-like, shape=(n_samples, n_features)]:
        Values where the acquisition function should be computed.

    * `model` [sklearn estimator that implements predict with ``return_std``]:
        The fit estimator that approximates the function through the
        method ``predict``.
        It should have a ``return_std`` parameter that returns the standard
        deviation.

    * `y_opt` [float, default 0]:
        Previous minimum value which we would like to improve upon.

    * `xi`: [float, default=0.01]:
        Controls how much improvement one wants over the previous best
        values. Useful only when ``method`` is set to "EI"

    * `return_grad`: [boolean, optional]:
        Whether or not to return the grad. Implemented only for the case where
        ``X`` is a single sample.

    Returns
    -------
    * `values`: [array-like, shape=(X.shape[0],)]:
        Acquisition function values computed at X.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        if return_grad:
            mu, std, mu_grad, std_grad = model.predict(
                X, return_std=True, return_mean_grad=True,
                return_std_grad=True)
        else:
            mu, std = model.predict(X, return_std=True)

    # check dimensionality of mu, std so we can divide them below
    if (mu.ndim != 1) or (std.ndim != 1):
        raise ValueError("mu and std are {}-dimensional and {}-dimensional, "
                         "however both must be 1-dimensional. Did you train "
                         "your model with an (N, 1) vector instead of an "
                         "(N,) vector?"
                         .format(mu.ndim, std.ndim))

    values = np.zeros_like(mu)
    mask = std > 0
    improve = y_opt - xi - mu[mask]
    scaled = improve / std[mask]
    values[mask] = norm.cdf(scaled)

    if return_grad:
        if not np.all(mask):
            return values, np.zeros_like(std_grad)

        # Substitute (y_opt - xi - mu) / sigma = t and apply chain rule.
        # improve_grad is the gradient of t wrt x.
        improve_grad = -mu_grad * std - std_grad * improve
        improve_grad /= std**2

        return values, improve_grad * norm.pdf(scaled)

    return values


def gaussian_ei(X, model, y_opt=0.0, xi=0.01, return_grad=False):
    """
    Use the expected improvement to calculate the acquisition values.

    The conditional probability `P(y=f(x) | x)`form a gaussian with a certain
    mean and standard deviation approximated by the model.

    The EI condition is derived by computing ``E[u(f(x))]``
    where ``u(f(x)) = 0``, if ``f(x) > y_opt`` and ``u(f(x)) = y_opt - f(x)``,
    if``f(x) < y_opt``.

    This solves one of the issues of the PI condition by giving a reward
    proportional to the amount of improvement got.

    Note that the value returned by this function should be maximized to
    obtain the ``X`` with maximum improvement.

    Parameters
    ----------
    * `X` [array-like, shape=(n_samples, n_features)]:
        Values where the acquisition function should be computed.

    * `model` [sklearn estimator that implements predict with ``return_std``]:
        The fit estimator that approximates the function through the
        method ``predict``.
        It should have a ``return_std`` parameter that returns the standard
        deviation.

    * `y_opt` [float, default 0]:
        Previous minimum value which we would like to improve upon.

    * `xi`: [float, default=0.01]:
        Controls how much improvement one wants over the previous best
        values. Useful only when ``method`` is set to "EI"

    * `return_grad`: [boolean, optional]:
        Whether or not to return the grad. Implemented only for the case where
        ``X`` is a single sample.

    Returns
    -------
    * `values`: [array-like, shape=(X.shape[0],)]:
        Acquisition function values computed at X.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        if return_grad:
            mu, std, mu_grad, std_grad = model.predict(
                X, return_std=True, return_mean_grad=True,
                return_std_grad=True)

        else:
            mu, std = model.predict(X, return_std=True)

    # check dimensionality of mu, std so we can divide them below
    if (mu.ndim != 1) or (std.ndim != 1):
        raise ValueError("mu and std are {}-dimensional and {}-dimensional, "
                         "however both must be 1-dimensional. Did you train "
                         "your model with an (N, 1) vector instead of an "
                         "(N,) vector?"
                         .format(mu.ndim, std.ndim))

    values = np.zeros_like(mu)
    mask = std > 0
    improve = y_opt - xi - mu[mask]
    scaled = improve / std[mask]
    cdf = norm.cdf(scaled)
    pdf = norm.pdf(scaled)
    exploit = improve * cdf
    explore = std[mask] * pdf
    values[mask] = exploit + explore

    if return_grad:
        if not np.all(mask):
            return values, np.zeros_like(std_grad)

        # Substitute (y_opt - xi - mu) / sigma = t and apply chain rule.
        # improve_grad is the gradient of t wrt x.
        improve_grad = -mu_grad * std - std_grad * improve
        improve_grad /= std ** 2
        cdf_grad = improve_grad * pdf
        pdf_grad = -improve * cdf_grad
        exploit_grad = -mu_grad * cdf - pdf_grad
        explore_grad = std_grad * pdf + pdf_grad

        grad = exploit_grad + explore_grad
        return values, grad

    return values



import warnings

try:
    from collections.abc import Sized
except ImportError:
    from collections import Sized
from collections import defaultdict
from functools import partial

import numpy as np
from scipy.stats import rankdata

import sklearn
from sklearn.base import is_classifier, clone
from sklearn.model_selection._search import BaseSearchCV
from sklearn.utils import check_random_state
from sklearn.utils.validation import indexable, check_is_fitted

try:
    from sklearn.metrics import check_scoring
except ImportError:
    from sklearn.metrics.scorer import check_scoring

from joblib import Parallel, delayed

from . import Optimizer
from .utils import point_asdict, dimensions_aslist, eval_callbacks
from .space import check_dimension
from .callbacks import check_callback


class BayesSearchCV(BaseSearchCV):
    """Bayesian optimization over hyper parameters.

    BayesSearchCV implements a "fit" and a "score" method.
    It also implements "predict", "predict_proba", "decision_function",
    "transform" and "inverse_transform" if they are implemented in the
    estimator used.

    The parameters of the estimator used to apply these methods are optimized
    by cross-validated search over parameter settings.

    In contrast to GridSearchCV, not all parameter values are tried out, but
    rather a fixed number of parameter settings is sampled from the specified
    distributions. The number of parameter settings that are tried is
    given by n_iter.

    Parameters are presented as a list of ProcessOptimizer.space.Dimension
    objects.

    Parameters
    ----------
    estimator : estimator object.
        A object of that type is instantiated for each search point.
        This object is assumed to implement the scikit-learn estimator api.
        Either estimator needs to provide a ``score`` function,
        or ``scoring`` must be passed.

    search_spaces : dict, list of dict or list of tuple containing
        (dict, int).
        One of these cases:
        1. dictionary, where keys are parameter names (strings)
        and values are ProcessOptimizer.space.Dimension instances (Real,
        Integer or Categorical) or any other valid value that defines
        ProcessOptimizer dimension (see ProcessOptimizer.Optimizer docs).
        Represents search space over parameters of the provided estimator.
        2. list of dictionaries: a list of dictionaries, where every
        dictionary fits the description given in case 1 above.
        If a list of dictionary objects is given, then the search is
        performed sequentially for every parameter space with maximum
        number of evaluations set to self.n_iter.
        3. list of (dict, int > 0): an extension of case 2 above,
        where first element of every tuple is a dictionary representing
        some search subspace, similarly as in case 2, and second element
        is a number of iterations that will be spent optimizing over
        this subspace.

    n_iter : int, default=50
        Number of parameter settings that are sampled. n_iter trades
        off runtime vs quality of the solution. Consider increasing
        ``n_points`` if you want to try more parameter settings in
        parallel.

    optimizer_kwargs : dict, optional
        Dict of arguments passed to :class:`Optimizer`.  For example,
        ``{'base_estimator': 'RF'}`` would use a Random Forest surrogate
        instead of the default Gaussian Process.

    scoring : string, callable or None, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.
        If ``None``, the ``score`` method of the estimator is used.

    fit_params : dict, optional
        Parameters to pass to the fit method.

    n_jobs : int, default=1
        Number of jobs to run in parallel. At maximum there are
        ``n_points`` times ``cv`` jobs available during each iteration.

    n_points : int, default=1
        Number of parameter settings to sample in parallel. If this does
        not align with ``n_iter``, the last iteration will sample less
        points. See also :func:`~Optimizer.ask`

    pre_dispatch : int, or string, optional
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:

            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs

            - An int, giving the exact number of total jobs that are
              spawned

            - A string, giving an expression as a function of n_jobs,
              as in '2*n_jobs'

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
          - None, to use the default 3-fold cross validation,
          - integer, to specify the number of folds in a `(Stratified)KFold`,
          - An object to be used as a cross-validation generator.
          - An iterable yielding train, test splits.

        For integer/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

    refit : boolean, default=True
        Refit the best estimator with the entire dataset.
        If "False", it is impossible to make predictions using
        this RandomizedSearchCV instance after fitting.

    verbose : integer
        Controls the verbosity: the higher, the more messages.

    random_state : int or RandomState
        Pseudo random number generator state used for random uniform sampling
        from lists of possible values instead of scipy.stats distributions.

    error_score : 'raise' (default) or numeric
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised. If a numeric value is given,
        FitFailedWarning is raised. This parameter does not affect the refit
        step, which will always raise the error.

    return_train_score : boolean, default=False
        If ``'True'``, the ``cv_results_`` attribute will include training
        scores.

    Example
    -------

    from ProcessOptimizer import BayesSearchCV
    # parameter ranges are specified by one of below
    from ProcessOptimizer.space import Real, Categorical, Integer

    from sklearn.datasets import load_iris
    from sklearn.svm import SVC
    from sklearn.model_selection import train_test_split

    X, y = load_iris(True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=0.75,
                                                        random_state=0)

    # log-uniform: understand as search over p = exp(x) by varying x
    opt = BayesSearchCV(
        SVC(),
        {
            'C': Real(1e-6, 1e+6, prior='log-uniform'),
            'gamma': Real(1e-6, 1e+1, prior='log-uniform'),
            'degree': Integer(1,8),
            'kernel': Categorical(['linear', 'poly', 'rbf']),
        },
        n_iter=32
    )

    # executes bayesian optimization
    opt.fit(X_train, y_train)

    # model can be saved, used for predictions or scoring
    print(opt.score(X_test, y_test))

    Attributes
    ----------
    cv_results_ : dict of numpy (masked) ndarrays
        A dict with keys as column headers and values as columns, that can be
        imported into a pandas ``DataFrame``.

        For instance the below given table

        +--------------+-------------+-------------------+---+---------------+
        | param_kernel | param_gamma | split0_test_score |...|rank_test_score|
        +==============+=============+===================+===+===============+
        |    'rbf'     |     0.1     |        0.8        |...|       2       |
        +--------------+-------------+-------------------+---+---------------+
        |    'rbf'     |     0.2     |        0.9        |...|       1       |
        +--------------+-------------+-------------------+---+---------------+
        |    'rbf'     |     0.3     |        0.7        |...|       1       |
        +--------------+-------------+-------------------+---+---------------+

        will be represented by a ``cv_results_`` dict of::

            {
            'param_kernel' : masked_array(data = ['rbf', 'rbf', 'rbf'],
                                          mask = False),
            'param_gamma'  : masked_array(data = [0.1 0.2 0.3], mask = False),
            'split0_test_score'  : [0.8, 0.9, 0.7],
            'split1_test_score'  : [0.82, 0.5, 0.7],
            'mean_test_score'    : [0.81, 0.7, 0.7],
            'std_test_score'     : [0.02, 0.2, 0.],
            'rank_test_score'    : [3, 1, 1],
            'split0_train_score' : [0.8, 0.9, 0.7],
            'split1_train_score' : [0.82, 0.5, 0.7],
            'mean_train_score'   : [0.81, 0.7, 0.7],
            'std_train_score'    : [0.03, 0.03, 0.04],
            'mean_fit_time'      : [0.73, 0.63, 0.43, 0.49],
            'std_fit_time'       : [0.01, 0.02, 0.01, 0.01],
            'mean_score_time'    : [0.007, 0.06, 0.04, 0.04],
            'std_score_time'     : [0.001, 0.002, 0.003, 0.005],
            'params' : [{'kernel' : 'rbf', 'gamma' : 0.1}, ...],
            }

        NOTE that the key ``'params'`` is used to store a list of parameter
        settings dict for all the parameter candidates.

        The ``mean_fit_time``, ``std_fit_time``, ``mean_score_time`` and
        ``std_score_time`` are all in seconds.

    best_estimator_ : estimator
        Estimator that was chosen by the search, i.e. estimator
        which gave highest score (or smallest loss if specified)
        on the left out data. Not available if refit=False.

    best_score_ : float
        Score of best_estimator on the left out data.

    best_params_ : dict
        Parameter setting that gave the best results on the hold out data.

    best_index_ : int
        The index (of the ``cv_results_`` arrays) which corresponds to the best
        candidate parameter setting.

        The dict at ``search.cv_results_['params'][search.best_index_]`` gives
        the parameter setting for the best model, that gives the highest
        mean score (``search.best_score_``).

    scorer_ : function
        Scorer function used on the held out data to choose the best
        parameters for the model.

    n_splits_ : int
        The number of cross-validation splits (folds/iterations).

    Notes
    -----
    The parameters selected are those that maximize the score of the held-out
    data, according to the scoring parameter.

    If `n_jobs` was set to a value higher than one, the data is copied for each
    parameter setting(and not `n_jobs` times). This is done for efficiency
    reasons if individual jobs take very little time, but may raise errors if
    the dataset is large and not enough memory is available.  A workaround in
    this case is to set `pre_dispatch`. Then, the memory is copied only
    `pre_dispatch` many times. A reasonable value for `pre_dispatch` is `2 *
    n_jobs`.

    See Also
    --------
    :class:`GridSearchCV`:
        Does exhaustive search over a grid of parameters.

    """

    def __init__(self, estimator, search_spaces, optimizer_kwargs=None,
                 n_iter=50, scoring=None, fit_params=None, n_jobs=1,
                 n_points=1, iid='deprecated', refit=True, cv=None, verbose=0,
                 pre_dispatch='2*n_jobs', random_state=None,
                 error_score='raise', return_train_score=False):

        self.search_spaces = search_spaces
        self.n_iter = n_iter
        self.n_points = n_points
        self.random_state = random_state
        self.optimizer_kwargs = optimizer_kwargs
        self._check_search_space(self.search_spaces)
        # Temporary fix for compatibility with sklearn 0.20 and 0.21
        # See scikit-optimize#762
        # To be consistent with sklearn 0.21+, fit_params should be deprecated
        # in the constructor and be passed in ``fit``.
        self.fit_params = fit_params

        if iid != "deprecated":
            warnings.warn("The `iid` parameter has been deprecated "
                          "and will be ignored.")
        self.iid = iid  # For sklearn repr pprint

        super(BayesSearchCV, self).__init__(
            estimator=estimator, scoring=scoring,
            n_jobs=n_jobs, refit=refit, cv=cv, verbose=verbose,
            pre_dispatch=pre_dispatch, error_score=error_score,
            return_train_score=return_train_score)

    def _check_search_space(self, search_space):
        """Checks whether the search space argument is correct"""

        if len(search_space) == 0:
            raise ValueError(
                "The search_spaces parameter should contain at least one"
                "non-empty search space, got %s" % search_space
            )

        # check if space is a single dict, convert to list if so
        if isinstance(search_space, dict):
            search_space = [search_space]

        # check if the structure of the space is proper
        if isinstance(search_space, list):
            # convert to just a list of dicts
            dicts_only = []

            # 1. check the case when a tuple of space, n_iter is provided
            for elem in search_space:
                if isinstance(elem, tuple):
                    if len(elem) != 2:
                        raise ValueError(
                            "All tuples in list of search spaces should have"
                            "length 2, and contain (dict, int), got %s" % elem
                        )
                    subspace, n_iter = elem

                    if (not isinstance(n_iter, int)) or n_iter < 0:
                        raise ValueError(
                            "Number of iterations in search space should be"
                            "positive integer, got %s in tuple %s " %
                            (n_iter, elem)
                        )

                    # save subspaces here for further checking
                    dicts_only.append(subspace)
                elif isinstance(elem, dict):
                    dicts_only.append(elem)
                else:
                    raise TypeError(
                        "A search space should be provided as a dict or"
                        "tuple (dict, int), got %s" % elem)

            # 2. check all the dicts for correctness of contents
            for subspace in dicts_only:
                for k, v in subspace.items():
                    check_dimension(v)
        else:
            raise TypeError(
                "Search space should be provided as a dict or list of dict,"
                "got %s" % search_space)

    # copied for compatibility with 0.19 sklearn from 0.18 BaseSearchCV
    @property
    def best_score_(self):
        check_is_fitted(self, 'cv_results_')
        return self.cv_results_['mean_test_score'][self.best_index_]

    # copied for compatibility with 0.19 sklearn from 0.18 BaseSearchCV
    @property
    def best_params_(self):
        check_is_fitted(self, 'cv_results_')
        return self.cv_results_['params'][self.best_index_]

    # copied for compatibility with 0.19 sklearn from 0.18 BaseSearchCV
    def _fit(self, X, y, groups, parameter_iterable):
        """
        Actual fitting,  performing the search over parameters.
        Taken from https://github.com/scikit-learn/scikit-learn/blob/0.18.X
                    .../sklearn/model_selection/_search.py
        """

        estimator = self.estimator
        cv = sklearn.model_selection._validation.check_cv(
            self.cv, y, classifier=is_classifier(estimator))
        self.scorer_ = check_scoring(
            self.estimator, scoring=self.scoring)

        X, y, groups = indexable(X, y, groups)
        n_splits = cv.get_n_splits(X, y, groups)
        if self.verbose > 0 and isinstance(parameter_iterable, Sized):
            n_candidates = len(parameter_iterable)
            print(
                "Fitting {0} folds for each of {1} candidates, totalling"
                " {2} fits".format(
                    n_splits, n_candidates, n_candidates * n_splits
                    )
                    )

        base_estimator = clone(self.estimator)
        pre_dispatch = self.pre_dispatch

        cv_iter = list(cv.split(X, y, groups))
        out = Parallel(
            n_jobs=self.n_jobs, verbose=self.verbose,
            pre_dispatch=pre_dispatch
        )(delayed(sklearn.model_selection._validation._fit_and_score)(
            clone(base_estimator),
            X, y, self.scorer_,
            train, test, self.verbose, parameters,
            fit_params=self.fit_params,
            return_train_score=self.return_train_score,
            return_n_test_samples=True,
            return_times=True, return_parameters=True,
            error_score=self.error_score
        )
            for parameters in parameter_iterable
            for train, test in cv_iter)

        # if one choose to see train score, "out" will contain train score info
        if self.return_train_score:
            (_fit_fail, train_scores, test_scores, test_sample_counts,
             fit_time, score_time, parameters) = zip(*[dic.values()
                                                       for dic in out])
        else:
            (_fit_fail, test_scores, test_sample_counts,
             fit_time, score_time, parameters) = zip(*[dic.values()
                                                       for dic in out])

        candidate_params = parameters[::n_splits]
        n_candidates = len(candidate_params)

        results = dict()

        def _store(key_name, array, splits=False, rank=False):
            """A small helper to store the scores/times to the cv_results_"""
            array = np.array(array, dtype=np.float64).reshape(n_candidates,
                                                              n_splits)
            if splits:
                for split_i in range(n_splits):
                    results["split%d_%s"
                            % (split_i, key_name)] = array[:, split_i]

            array_means = np.average(array, axis=1)
            results['mean_%s' % key_name] = array_means
            # Weighted std is not directly available in numpy
            array_stds = np.sqrt(np.average((array -
                                             array_means[:, np.newaxis]) ** 2,
                                            axis=1))
            results['std_%s' % key_name] = array_stds

            if rank:
                results["rank_%s" % key_name] = np.asarray(
                    rankdata(-array_means, method='min'), dtype=np.int32)

        _store('test_score', test_scores, splits=True, rank=True)
        if self.return_train_score:
            _store('train_score', train_scores, splits=True)
        _store('fit_time', fit_time)
        _store('score_time', score_time)

        best_index = np.flatnonzero(results["rank_test_score"] == 1)[0]
        best_parameters = candidate_params[best_index]

        # Use one MaskedArray and mask all the places where the param is not
        # applicable for that candidate. Use defaultdict as each candidate may
        # not contain all the params
        param_results = defaultdict(partial(
            np.ma.array,
            np.empty(n_candidates,),
            mask=True,
            dtype=object))
        for cand_i, params in enumerate(candidate_params):
            for name, value in params.items():
                # An all masked empty array gets created for the key
                # `"param_%s" % name` at the first occurence of `name`.
                # Setting the value at an index also unmasks that index
                param_results["param_%s" % name][cand_i] = value

        results.update(param_results)

        # Store a list of param dicts at the key 'params'
        results['params'] = candidate_params

        self.cv_results_ = results
        self.best_index_ = best_index
        self.n_splits_ = n_splits

        if self.refit:
            # fit the best estimator using the entire dataset
            # clone first to work around broken estimators
            best_estimator = clone(base_estimator).set_params(
                **best_parameters)
            if y is not None:
                best_estimator.fit(X, y, **self.fit_params)
            else:
                best_estimator.fit(X, **self.fit_params)
            self.best_estimator_ = best_estimator
        return self

    def _fit_best_model(self, X, y):
        """Fit the estimator copy with best parameters found to the
        provided data.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples] or [n_samples, n_output],
            Target relative to X for classification or regression.

        Returns
        -------
        self
        """
        self.best_estimator_ = clone(self.estimator)
        self.best_estimator_.set_params(**self.best_params_)
        self.best_estimator_.fit(X, y, **(self.fit_params or {}))
        return self

    def _make_optimizer(self, params_space):
        """Instantiate ProcessOptimizer Optimizer class.

        Parameters
        ----------
        params_space : dict
            Represents parameter search space. The keys are parameter
            names (strings) and values are ProcessOptimizer.space.Dimension
            instances, one of Real, Integer or Categorical.

        Returns
        -------
        optimizer: Instance of the `Optimizer` class used for for search
            in some parameter space.

        """

        kwargs = self.optimizer_kwargs_.copy()
        kwargs['dimensions'] = dimensions_aslist(params_space)
        optimizer = Optimizer(**kwargs)

        return optimizer

    def _step(self, X, y, search_space, optimizer, groups=None, n_points=1):
        """Generate n_jobs parameters and evaluate them in parallel.
        """

        # get parameter values to evaluate
        params = optimizer.ask(n_points=n_points)

        # convert parameters to python native types
        if n_points == 1:
            params = [[np.array(v).item() for v in params]]
        else:
            params = [[np.array(v).item() for v in p] for p in params]

        # make lists into dictionaries
        params_dict = [point_asdict(search_space, p) for p in params]

        # HACK: self.cv_results_ is reset at every call to _fit, keep current
        all_cv_results = self.cv_results_

        # HACK: this adds compatibility with different versions of sklearn
        refit = self.refit
        self.refit = False
        self._fit(X, y, groups, params_dict)
        self.refit = refit

        # merge existing and new cv_results_
        for k in self.cv_results_:
            all_cv_results[k].extend(self.cv_results_[k])

        self.cv_results_ = all_cv_results
        self.best_index_ = np.argmax(self.cv_results_['mean_test_score'])

        # feed the point and objective back into optimizer
        local_results = self.cv_results_['mean_test_score'][-len(params):]

        # optimizer minimizes objective, hence provide negative score
        return optimizer.tell(params, [-score for score in local_results])

    @property
    def total_iterations(self):
        """
        Count total iterations that will be taken to explore
        all subspaces with `fit` method.

        Returns
        -------
        max_iter: int, total number of iterations to explore
        """
        total_iter = 0

        for elem in self.search_spaces:

            if isinstance(elem, tuple):
                space, n_iter = elem
            else:
                n_iter = self.n_iter

            total_iter += n_iter

        return total_iter

    def _run_search(self, x):
        pass

    def fit(self, X, y=None, groups=None, callback=None):
        """Run fit on the estimator with randomly drawn parameters.

        Parameters
        ----------
        X : array-like or sparse matrix, shape = [n_samples, n_features]
            The training input samples.

        y : array-like, shape = [n_samples] or [n_samples, n_output]
            Target relative to X for classification or regression (class
            labels should be integers or strings).

        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set.

        callback: [callable, list of callables, optional]
            If callable then `callback(res)` is called after each parameter
            combination tested. If list of callables, then each callable in
            the list is called.
        """

        # check if space is a single dict, convert to list if so
        search_spaces = self.search_spaces
        if isinstance(search_spaces, dict):
            search_spaces = [search_spaces]

        callbacks = check_callback(callback)

        if self.optimizer_kwargs is None:
            self.optimizer_kwargs_ = {}
        else:
            self.optimizer_kwargs_ = dict(self.optimizer_kwargs)
        random_state = check_random_state(self.random_state)
        self.optimizer_kwargs_['random_state'] = random_state

        # Instantiate optimizers for all the search spaces.
        optimizers = []
        for search_space in search_spaces:
            if isinstance(search_space, tuple):
                search_space = search_space[0]
            optimizers.append(self._make_optimizer(search_space))
        self.optimizers_ = optimizers  # will save the states of the optimizers

        self.cv_results_ = defaultdict(list)
        self.best_index_ = None
        self.multimetric_ = False
        self.optimizer_results_ = {}

        n_points = self.n_points

        for search_space, optimizer in zip(search_spaces, optimizers):
            # if not provided with search subspace, n_iter is taken as
            # self.n_iter
            if isinstance(search_space, tuple):
                search_space, n_iter = search_space
            else:
                n_iter = self.n_iter

            # do the optimization for particular search space
            while n_iter > 0:
                # when n_iter < n_points points left for evaluation
                n_points_adjusted = min(n_iter, n_points)

                optim_result = self._step(
                    X, y, search_space, optimizer,
                    groups=groups, n_points=n_points_adjusted
                )
                n_iter -= n_points

                if eval_callbacks(callbacks, optim_result):
                    break

            self.optimizer_results_[optimizer] = optim_result

        # Refit the best model on the the whole dataset
        if self.refit:
            self._fit_best_model(X, y)

        return self



"""Random search."""

from .base import base_minimize


def dummy_minimize(func, dimensions, n_calls=100, x0=None, y0=None,
                   random_state=None, verbose=False, callback=None):
    """Random search by uniform sampling within the given bounds.

    Parameters
    ----------
    * `func` [callable]:
        Function to minimize. Should take a single list of parameters
        and return the objective value.
    
        If you have a search-space where all dimensions have names,
        then you can use `ProcessOptimizer.utils.use_named_args` as a decorator
        on your objective function, in order to call it directly
        with the named arguments. See `use_named_args` for an example.

    * `dimensions` [list, shape=(n_dims,)]:
        List of search space dimensions.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, prior)` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

    * `n_calls` [int, default=100]:
        Number of calls to `func` to find the minimum.

    * `x0` [list, list of lists or `None`]:
        Initial input points.

        - If it is a list of lists, use it as a list of input points.
        - If it is a list, use it as a single initial input point.
        - If it is `None`, no initial input points are used.

    * `y0` [list, scalar or `None`]:
        Evaluation of initial input points.

        - If it is a list, then it corresponds to evaluations of the function
          at each element of `x0` : the i-th element of `y0` corresponds
          to the function evaluated at the i-th element of `x0`.
        - If it is a scalar, then it corresponds to the evaluation of the
          function at `x0`.
        - If it is None and `x0` is provided, then the function is evaluated
          at each element of `x0`.

    * `random_state` [int, RandomState instance, or None (default)]:
        Set random state to something other than None for reproducible
        results.

    * `verbose` [boolean, default=False]:
        Control the verbosity. It is advised to set the verbosity to True
        for long optimization runs.

    * `callback` [callable, list of callables, optional]
        If callable then `callback(res)` is called after each call to `func`.
        If list of callables, then each callable in the list is called.

    Returns
    -------
    * `res` [`OptimizeResult`, scipy object]:
        The optimization result returned as a OptimizeResult object.
        Important attributes are:

        - `x` [list]: location of the minimum.
        - `fun` [float]: function value at the minimum.
        - `x_iters` [list of lists]: location of function evaluation for each
           iteration.
        - `func_vals` [array]: function value for each iteration.
        - `space` [Space]: the optimisation space.
        - `specs` [dict]: the call specifications.
        - `rng` [RandomState instance]: State of the random state
           at the end of minimization.

        For more details related to the OptimizeResult object, refer
        http://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html
    """
    # all our calls want random suggestions, except if we need to evaluate
    # some initial points
    if x0 is not None and y0 is None:
      if isinstance(x0[0], (list, tuple)):
        n_random_calls = n_calls - len(x0)
      else:
        n_random_calls = n_calls - len([x0])
    else:
      n_random_calls = n_calls

    return base_minimize(func, dimensions, base_estimator="dummy",
                         # explicitly set optimizer to sampling as "dummy"
                         # minimizer does not provide gradients.
                         acq_optimizer="sampling",
                         n_calls=n_calls, n_random_starts=n_random_calls,
                         x0=x0, y0=y0, random_state=random_state,
                         verbose=verbose,
                         callback=callback)



from math import sqrt

import numpy as np
from sklearn.gaussian_process.kernels import Kernel as sk_Kernel
from sklearn.gaussian_process.kernels import ConstantKernel as sk_ConstantKernel
from sklearn.gaussian_process.kernels import DotProduct as sk_DotProduct
from sklearn.gaussian_process.kernels import Exponentiation as sk_Exponentiation
from sklearn.gaussian_process.kernels import ExpSineSquared as sk_ExpSineSquared
from sklearn.gaussian_process.kernels import Hyperparameter
from sklearn.gaussian_process.kernels import Matern as sk_Matern
from sklearn.gaussian_process.kernels import NormalizedKernelMixin as sk_NormalizedKernelMixin
from sklearn.gaussian_process.kernels import Product as sk_Product
from sklearn.gaussian_process.kernels import RationalQuadratic as sk_RationalQuadratic
from sklearn.gaussian_process.kernels import RBF as sk_RBF
from sklearn.gaussian_process.kernels import StationaryKernelMixin as sk_StationaryKernelMixin
from sklearn.gaussian_process.kernels import Sum as sk_Sum
from sklearn.gaussian_process.kernels import WhiteKernel as sk_WhiteKernel


class Kernel(sk_Kernel):
    """
    Base class for ProcessOptimizer.gaussian_process kernels.
    Supports computation of the gradient of the kernel with respect to X
    """
    def __add__(self, b):
        if not isinstance(b, Kernel):
            return Sum(self, ConstantKernel(b))
        return Sum(self, b)

    def __radd__(self, b):
        if not isinstance(b, Kernel):
            return Sum(ConstantKernel(b), self)
        return Sum(b, self)

    def __mul__(self, b):
        if not isinstance(b, Kernel):
            return Product(self, ConstantKernel(b))
        return Product(self, b)

    def __rmul__(self, b):
        if not isinstance(b, Kernel):
            return Product(ConstantKernel(b), self)
        return Product(b, self)

    def __pow__(self, b):
        return Exponentiation(self, b)

    def gradient_x(self, x, X_train):
        """
        Computes gradient of K(x, X_train) with respect to x

        Parameters
        ----------
        x: array-like, shape=(n_features,)
            A single test point.

        Y: array-like, shape=(n_samples, n_features)
            Training data used to fit the gaussian process.

        Returns
        -------
        gradient_x: array-like, shape=(n_samples, n_features)
            Gradient of K(x, X_train) with respect to x.
        """
        raise NotImplementedError


class RBF(Kernel, sk_RBF):
    def gradient_x(self, x, X_train):
        # diff = (x - X) / length_scale
        # size = (n_train_samples, n_dimensions)
        x = np.asarray(x)
        X_train = np.asarray(X_train)

        length_scale = np.asarray(self.length_scale)
        diff = x - X_train
        diff /= length_scale

        # e = -exp(0.5 * \sum_{i=1}^d (diff ** 2))
        # size = (n_train_samples, 1)
        exp_diff_squared = np.sum(diff**2, axis=1)
        exp_diff_squared *= -0.5
        exp_diff_squared = np.exp(exp_diff_squared, exp_diff_squared)
        exp_diff_squared = np.expand_dims(exp_diff_squared, axis=1)
        exp_diff_squared *= -1

        # gradient = (e * diff) / length_scale
        gradient = exp_diff_squared * diff
        gradient /= length_scale
        return gradient


class Matern(Kernel, sk_Matern):
    def gradient_x(self, x, X_train):
        x = np.asarray(x)
        X_train = np.asarray(X_train)
        length_scale = np.asarray(self.length_scale)

        # diff = (x - X_train) / length_scale
        # size = (n_train_samples, n_dimensions)
        diff = x - X_train
        diff /= length_scale

        # dist_sq = \sum_{i=1}^d (diff ^ 2)
        # dist = sqrt(dist_sq)
        # size = (n_train_samples,)
        dist_sq = np.sum(diff**2, axis=1)
        dist = np.sqrt(dist_sq)

        if self.nu == 0.5:
            # e = -np.exp(-dist) / dist
            # size = (n_train_samples, 1)
            scaled_exp_dist = -dist
            scaled_exp_dist = np.exp(scaled_exp_dist, scaled_exp_dist)
            scaled_exp_dist *= -1

            # grad = (e * diff) / length_scale
            # For all i in [0, D) if x_i equals y_i.
            # 1. e -> -1
            # 2. (x_i - y_i) / \sum_{j=1}^D (x_i - y_i)**2 approaches 1.
            # Hence the gradient when for all i in [0, D),
            # x_i equals y_i is -1 / length_scale[i].
            gradient = -np.ones((X_train.shape[0], x.shape[0]))
            mask = dist != 0.0
            scaled_exp_dist[mask] /= dist[mask]
            scaled_exp_dist = np.expand_dims(scaled_exp_dist, axis=1)
            gradient[mask] = scaled_exp_dist[mask] * diff[mask]
            gradient /= length_scale
            return gradient

        elif self.nu == 1.5:
            # grad(fg) = f'g + fg'
            # where f = 1 + sqrt(3) * euclidean((X - Y) / length_scale)
            # where g = exp(-sqrt(3) * euclidean((X - Y) / length_scale))
            sqrt_3_dist = sqrt(3) * dist
            f = np.expand_dims(1 + sqrt_3_dist, axis=1)

            # When all of x_i equals y_i, f equals 1.0, (1 - f) equals
            # zero, hence from below
            # f * g_grad + g * f_grad (where g_grad = -g * f_grad)
            # -f * g * f_grad + g * f_grad
            # g * f_grad * (1 - f) equals zero.
            # sqrt_3_by_dist can be set to any value since diff equals
            # zero for this corner case.
            sqrt_3_by_dist = np.zeros_like(dist)
            nzd = dist != 0.0
            sqrt_3_by_dist[nzd] = sqrt(3) / dist[nzd]
            dist_expand = np.expand_dims(sqrt_3_by_dist, axis=1)

            f_grad = diff / length_scale
            f_grad *= dist_expand

            sqrt_3_dist *= -1
            exp_sqrt_3_dist = np.exp(sqrt_3_dist, sqrt_3_dist)
            g = np.expand_dims(exp_sqrt_3_dist, axis=1)
            g_grad = -g * f_grad

            # f * g_grad + g * f_grad (where g_grad = -g * f_grad)
            f *= -1
            f += 1
            return g * f_grad * f

        elif self.nu == 2.5:
            # grad(fg) = f'g + fg'
            # where f = (1 + sqrt(5) * euclidean((X - Y) / length_scale) +
            #            5 / 3 * sqeuclidean((X - Y) / length_scale))
            # where g = exp(-sqrt(5) * euclidean((X - Y) / length_scale))
            sqrt_5_dist = sqrt(5) * dist
            f2 = (5.0 / 3.0) * dist_sq
            f2 += sqrt_5_dist
            f2 += 1
            f = np.expand_dims(f2, axis=1)

            # For i in [0, D) if x_i equals y_i
            # f = 1 and g = 1
            # Grad = f'g + fg' = f' + g'
            # f' = f_1' + f_2'
            # Also g' = -g * f1'
            # Grad = f'g - g * f1' * f
            # Grad = g * (f' - f1' * f)
            # Grad = f' - f1'
            # Grad = f2' which equals zero when x = y
            # Since for this corner case, diff equals zero,
            # dist can be set to anything.
            nzd_mask = dist != 0.0
            nzd = dist[nzd_mask]
            dist[nzd_mask] = np.reciprocal(nzd, nzd)

            dist *= sqrt(5)
            dist = np.expand_dims(dist, axis=1)
            diff /= length_scale
            f1_grad = dist * diff
            f2_grad = (10.0 / 3.0) * diff
            f_grad = f1_grad + f2_grad

            sqrt_5_dist *= -1
            g = np.exp(sqrt_5_dist, sqrt_5_dist)
            g = np.expand_dims(g, axis=1)
            g_grad = -g * f1_grad
            return f * g_grad + g * f_grad


class RationalQuadratic(Kernel, sk_RationalQuadratic):

    def gradient_x(self, x, X_train):
        x = np.asarray(x)
        X_train = np.asarray(X_train)
        alpha = self.alpha
        length_scale = self.length_scale

        # diff = (x - X_train) / length_scale
        # size = (n_train_samples, n_dimensions)
        diff = x - X_train
        diff /= length_scale

        # dist = -(1 + (\sum_{i=1}^d (diff^2) / (2 * alpha)))** (-alpha - 1)
        # size = (n_train_samples,)
        scaled_dist = np.sum(diff**2, axis=1)
        scaled_dist /= (2 * self.alpha)
        scaled_dist += 1
        scaled_dist **= (-alpha - 1)
        scaled_dist *= -1

        scaled_dist = np.expand_dims(scaled_dist, axis=1)
        diff_by_ls = diff / length_scale
        return scaled_dist * diff_by_ls


class ExpSineSquared(Kernel, sk_ExpSineSquared):

    def gradient_x(self, x, X_train):
        x = np.asarray(x)
        X_train = np.asarray(X_train)
        length_scale = self.length_scale
        periodicity = self.periodicity

        diff = x - X_train
        sq_dist = np.sum(diff**2, axis=1)
        dist = np.sqrt(sq_dist)

        pi_by_period = dist * (np.pi / periodicity)
        sine = np.sin(pi_by_period) / length_scale
        sine_squared = -2 * sine**2
        exp_sine_squared = np.exp(sine_squared)

        grad_wrt_exp = -2 * np.sin(2 * pi_by_period) / length_scale**2

        # When x_i -> y_i for all i in [0, D), the gradient becomes
        # zero. See https://github.com/MechCoder/Notebooks/blob/master/ExpSineSquared%20Kernel%20gradient%20computation.ipynb
        # for a detailed math explanation
        # grad_wrt_theta can be anything since diff is zero
        # for this corner case, hence we set to zero.
        grad_wrt_theta = np.zeros_like(dist)
        nzd = dist != 0.0
        grad_wrt_theta[nzd] = np.pi / (periodicity * dist[nzd])
        return np.expand_dims(
            grad_wrt_theta * exp_sine_squared * grad_wrt_exp, axis=1) * diff


class ConstantKernel(Kernel, sk_ConstantKernel):

    def gradient_x(self, x, X_train):
        return np.zeros_like(X_train)


class WhiteKernel(Kernel, sk_WhiteKernel):

    def gradient_x(self, x, X_train):
        return np.zeros_like(X_train)


class Exponentiation(Kernel, sk_Exponentiation):

    def gradient_x(self, x, X_train):
        x = np.asarray(x)
        X_train = np.asarray(X_train)
        expo = self.exponent
        kernel = self.kernel

        K = np.expand_dims(
            kernel(np.expand_dims(x, axis=0), X_train)[0], axis=1)
        return expo * K ** (expo - 1) * kernel.gradient_x(x, X_train)


class Sum(Kernel, sk_Sum):

    def gradient_x(self, x, X_train):
        return (
            self.k1.gradient_x(x, X_train) +
            self.k2.gradient_x(x, X_train)
        )


class Product(Kernel, sk_Product):

    def gradient_x(self, x, X_train):
        x = np.asarray(x)
        x = np.expand_dims(x, axis=0)
        X_train = np.asarray(X_train)
        f_ggrad = (
            np.expand_dims(self.k1(x, X_train)[0], axis=1) *
            self.k2.gradient_x(x, X_train)
        )
        fgrad_g = (
            np.expand_dims(self.k2(x, X_train)[0], axis=1) *
            self.k1.gradient_x(x, X_train)
        )
        return f_ggrad + fgrad_g


class DotProduct(Kernel, sk_DotProduct):

    def gradient_x(self, x, X_train):
        return np.asarray(X_train)


class HammingKernel(sk_StationaryKernelMixin, sk_NormalizedKernelMixin,
                    Kernel):
    """
    The HammingKernel is used to handle categorical inputs.

    ``K(x_1, x_2) = exp(\sum_{j=1}^{d} -ls_j * (I(x_1j != x_2j)))``

    Parameters
    -----------
    * `length_scale` [float, array-like, shape=[n_features,], 1.0 (default)]
        The length scale of the kernel. If a float, an isotropic kernel is
        used. If an array, an anisotropic kernel is used where each dimension
        of l defines the length-scale of the respective feature dimension.

    * `length_scale_bounds` [array-like, [1e-5, 1e5] (default)]
        The lower and upper bound on length_scale
    """

    def __init__(self, length_scale=1.0, length_scale_bounds=(1e-5, 1e5)):
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds

    @property
    def hyperparameter_length_scale(self):
        length_scale = self.length_scale
        anisotropic = np.iterable(length_scale) and len(length_scale) > 1
        if anisotropic:
            return Hyperparameter("length_scale", "numeric",
                                  self.length_scale_bounds,
                                  len(length_scale))
        return Hyperparameter(
            "length_scale", "numeric", self.length_scale_bounds)

    def __call__(self, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        * `X` [array-like, shape=(n_samples_X, n_features)]
            Left argument of the returned kernel k(X, Y)

        * `Y` [array-like, shape=(n_samples_Y, n_features) or None(default)]
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.

        * `eval_gradient` [bool, False(default)]
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.

        Returns
        -------
        * `K` [array-like, shape=(n_samples_X, n_samples_Y)]
            Kernel k(X, Y)

        * `K_gradient` [array-like, shape=(n_samples_X, n_samples_X, n_dims)]
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """
        length_scale = self.length_scale
        anisotropic = np.iterable(length_scale) and len(length_scale) > 1

        if np.iterable(length_scale):
            if len(length_scale) > 1:
                length_scale = np.asarray(length_scale, dtype=float)
            else:
                length_scale = float(length_scale[0])
        else:
            length_scale = float(length_scale)

        X = np.atleast_2d(X)
        if anisotropic and X.shape[1] != len(length_scale):
            raise ValueError(
                "Expected X to have %d features, got %d" %
                (len(length_scale), X.shape[1]))

        n_samples, n_dim = X.shape

        Y_is_None = Y is None
        if Y_is_None:
            Y = X
        elif eval_gradient:
            raise ValueError("gradient can be evaluated only when Y != X")
        else:
            Y = np.atleast_2d(Y)

        indicator = np.expand_dims(X, axis=1) != Y
        kernel_prod = np.exp(-np.sum(length_scale * indicator, axis=2))

        # dK / d theta = (dK / dl) * (dl / d theta)
        # theta = log(l) => dl / d (theta) = e^theta = l
        # dK / d theta = l * dK / dl

        # dK / dL computation
        if anisotropic:
            grad = (-np.expand_dims(kernel_prod, axis=-1) *
                    np.array(indicator, dtype=np.float32))
        else:
            grad = -np.expand_dims(kernel_prod * np.sum(indicator, axis=2),
                                   axis=-1)

        grad *= length_scale
        if eval_gradient:
            return kernel_prod, grad
        return kernel_prod



import numpy as np
from ProcessOptimizer import expected_minimum

class ModelSystem:
    """
    Model System for testing the ProcessOptimizer. Instances of this class 
    will be used in the example notebooks. 

    Parameters:
    * `score` [callable]:
        Function for calculating the score of the system at a given point in 
        the parameter space. It is expected to have the following signature:

            def score(x, rng=np.random.default_rng(), noise_std=noise_value):
                # * x is the point in the parameter space where the score will 
                #   be evaluated. Note that x is in the original coordinates of
                #   the parameter space, not normalized coordinates. 
                # * rng is the random number generator used for adding noise 
                #   to the system. The user can set a seed through this 
                #   parameter. There must be a default value for this parameter. 
                # * noise_std is the standard deviation of the noise. There
                #   must be a default value for this parameter. 
                ....
                # The score of the system is returned
                return score

    * `space` [Space]:
        The parameter space in the form of a Space object. 

    * `true_min` [float]:
        The true minimum value of the score function within the parameter space. 
    """
    def __init__(self, score, space, true_min=None):
        self.space = space
        self.score = score
        if true_min is None:
            ndims = space.n_dims()
            points = space.lhs(ndims*10)
            scores = [score(point) for point in points]
            true_min = np.min(scores)
        self.true_min = true_min
        
    def result_loss(self, result):
        """Calculate the loss of the optimization result. 

        Parameters:
        * `result` [OptimizeResult object of scipy.optimize.optimize module]:
            The result of an optimization. 

        Returns
        * loss [float]:
            The loss of the system, i.e. the difference between the true system 
            value at the location of the model's expected minimum and the best 
            possible system value. 
        """
        # Get the location of the expected minimum
        model_x,_ = expected_minimum(result)
        # Calculate the difference between the score at model_x and the true minimum value
        loss = self.score(model_x, noise_std=0) - self.true_min
        return loss


from .base import base_minimize
from .dummy import dummy_minimize
from .forest import forest_minimize
from .gbrt import gbrt_minimize
from .gp import gp_minimize
from .optimizer import Optimizer


__all__ = [
    "base_minimize", "dummy_minimize",
    "forest_minimize", "gbrt_minimize", "gp_minimize",
    "Optimizer"
]




from bokeh.layouts import row, column
from bokeh.models import ColumnDataSource, Button, Span, CustomJS
from bokeh.models.widgets import Slider, Toggle
from bokeh.models.widgets import CheckboxButtonGroup, Div, Select
from bokeh.plotting import figure

from ProcessOptimizer.space import Categorical
from ProcessOptimizer.plots import _map_categories, dependence
from ProcessOptimizer import expected_minimum, expected_minimum_random_sampling
import numpy as np

import matplotlib.pyplot as plt
import math
import copy
from bokeh.server.server import Server
import matplotlib as mpl

"""
Quick introduction.
The GUI conists of several elements.
Plots: the plots as they also appear in objective_plot
Selectors: Sliders that appear on top of the plots with which the red markers
can be adjusted
Toggle_x button group: A group of buttons where we can toggle which parameters
we want to plot.
Generate button: Dependence is only calculated when this buton is pressed
Other buttons: With buttons we can toggle partial dependence on and off,
control the n_points variable (resolution when calculating dependence)
and control the evaluation method
Y values: A small text field which shows the expected value
as well as confidence interval
"""

# Here we define the global variables
result = None
# x_eval_selectors holds the sliders and selectors for parameter values
x_eval_selectors = None
# old_active_list determines which x_eval_selectors were present
# before last update
old_active_list = None
# The 'reds' field is a (NxN) list. The diagonal holds a Span
source = {'reds': []}
# x_eval_selectors_values is a "location" value,
# that determines the location for the red line in the 1d plots.
# The off diagonals
# holds coordinates for the red markers used in the contour plot.
x_eval_selectors_values = None
# max_pars defines the maximum number of parameters that can be plotted
max_pars = None
layout = None
button_partial_dependence = None
button_color_map = None
button_draw_confidence = None
button_generate = None
buttons_toggle_x = None
dropdown_eval_method = None
slider_n_points = None
y_value = None
colorbar = None


def set_globals(parsed_result):
    global button_generate, buttons_toggle_x, dropdown_eval_method
    global slider_n_points, max_pars, source, old_active_list
    global result, x_eval_selectors, layout, x_eval_selectors_values
    global button_partial_dependence, y_value, button_color_map, colorbar
    global button_draw_confidence

    # Here we set the values of the global variables and create the layout
    result = parsed_result
    x_eval_selectors_values = copy.copy(result.x)
    max_pars = len(result.space.dimensions)

    # Layout
    button_generate = Button(label="Generate", button_type="success")
    button_generate.on_click(lambda: handle_button_generate(layout, result))
    buttons_toggle_x = CheckboxButtonGroup(
        labels=['x '+str(s) for s in range(max_pars)], active=[])
    button_partial_dependence = Toggle(
        label="Use partial dependence", button_type="default")
    button_color_map = Toggle(
        label="Use same color map", button_type="default")
    button_draw_confidence = Toggle(
        label="Draw upper confidence limit", button_type="default")
    dropdown_eval_method = Select(title="Evaluation method:",
                                  value='Result',
                                  options=['Result',
                                           'Exp min',
                                           'Exp min rand',
                                           'Sliders'],
                                  width=200, height=40
                                  )
    slider_n_points = Slider(start=1, end=40, value=20,
                             step=1, title="n-points", width=200, height=10)
    y_value = Div(text="""""", width=300, height=200)
    # We initiate colorbar as an empty div first
    # and then change it to a plot in case we want to show it
    colorbar = Div()
    row_plots = row([])
    row_top = row(button_generate, buttons_toggle_x)
    col_right_side = column(button_partial_dependence,
                            button_color_map,
                            button_draw_confidence,
                            dropdown_eval_method,
                            slider_n_points,
                            y_value,
                            colorbar
                            )
    col_left_side = column(row_top, row_plots)
    layout = row(col_left_side, col_right_side)


def handle_button_generate(layout, result):
    global old_active_list
    # Remove plots when button is clicked.
    # If we dont do this we ge the "Cant apply patch" bug.
    layout.children[0].children[1] = Div(text="""""",
                                         width=500, height=100)
    # Callback for when generate button gets pressed
    active_list = get_active_list()  # Get the current active list
    n_points = get_n_points()

    # Updating plots
    if active_list:  # Only plot if there is at least one selection
        # x_eval is used both for red markers and as the values for
        x_eval = get_x_eval(result, active_list)
        x_eval_transformed = result.space.transform([x_eval])
        # Calculate y and confidence interval
        y, std = result.models[-1].predict(x_eval_transformed, return_std=True)
        confidence = [round(y[0]-1.96*std[0], 5), round(y[0]+1.96*std[0], 5)]
        # We update the part of the layout that contains the plots
        plots, colorbar_interval = get_plots_layout(
            layout, result, active_list, n_points, x_eval, confidence)

        layout.children[0].children[1] = plots
        # Update text in right side of GUI
        y_value.text = """<font size="5"><b><br>"""+'Y = ' + \
            str(round(y[0], 5))+'</b><br> (' + str(confidence[0]) + \
            ', '+str(confidence[1]) + ')'+'</font>'
        if colorbar_interval:
            colorbar_plot = figure(plot_height=200,
                                   plot_width=100,
                                   tools='',
                                   x_range=[0, 1],
                                   y_range=colorbar_interval)
            im = get_colorbar_as_rgba()
            colorbar_plot.image_rgba(
                image=[im],
                x=0,
                y=colorbar_interval[0],
                dw=1,
                dh=colorbar_interval[1]-colorbar_interval[0])
            colorbar_plot.xaxis.visible = False
            colorbar_plot.toolbar.logo = None
        else:
            colorbar_plot = Div()
        layout.children[1].children[6] = colorbar_plot
    else:  # If no selection we encourage the user to select some parameters
        layout.children[0].children[1] = Div(
            text="""<font size="10"><br><br>Let's select
            som parameters to plot!</font>""",
            width=500,
            height=100)
    # Update the old_active_list for next update
    old_active_list = active_list


def get_x_eval_selectors_values():
    """Description of x_eval_selectors_values

    Returns the values for the x_eval_selectors. Uses the global
    x_eval_selectors_values and for each selector that is present in the GUI
    we replace the corresponding value. Even though a selector isnot present
    in the GUI the last value that it had i still retrived.

    """
    global x_eval_selectors
    global old_active_list
    values = x_eval_selectors_values
    if True:
        n = 0  # The index of the selctors in GUI
        for i in old_active_list:  # Use value from GUI selector instead
            val = x_eval_selectors[n].value
            values[i] = val
            n += 1
    values = x_eval_selectors_values
    return values


def get_active_list():
    """Description of get_active_list

    Returns the a list of the indexes of parameters that have been toggled in
    the buttons_toggle_x button-group in the GUI

    """

    return buttons_toggle_x.active


def get_n_points():
    # Returns the value of the slider_n_points slider in the GUI
    return slider_n_points.value


def get_use_partial_dependence():
    """Description of get_use_partial_dependence

    Returns True or false depending on wether or not the partial dependence
    button is toggled in the GUI

    """

    return button_partial_dependence.active


def get_use_same_color_map():

    return button_color_map.active


def get_plot_list(layout, result, active_list, n_points, x_eval, confidence):
    """Description of get_plot_list

    Returns a NxN list of plots where N is the number of parameters to be
    plotted. The diagonal is 1d plots and the off diagonal is contour plots.

    """

    global source
    use_same_color_map = get_use_same_color_map()
    draw_confidence = button_draw_confidence.active
    if get_use_partial_dependence():
        # Not passing any eval values to dependency function
        # makes it calculate partial dependence
        dependence_eval = None
    else:
        dependence_eval = x_eval

    # First we calculate all the data that will be used for all the plots.
    # After this we draw the plots.
    plots_data = []
    # We need to know the maxmimum and minimum values before drawing the plots,
    # in order to create a correct color-mapping
    plots = []
    space = result.space
    model = result.models[-1]
    bounds = result.space.bounds
    # the iscat variable is a list of bools. True if parameter is categorical.
    # False otherwise. red_vals are the coordinates for the x_eval
    # that is mapped to integers in case they are categorical.
    _, red_vals, iscat = _map_categories(space, result.x_iters, x_eval)
    # We add 0.5 to the value of all categorical integers.
    # This is due to the way bokeh handles plotting of categorical values.
    red_vals = [val+0.5 if iscat[i] else val for i, val in enumerate(red_vals)]
    # While iterating through each plot
    # we keep track of the max and min values of y and z in case of 2d plots.
    # We use these values to set the axis of all plots,
    # to make them the same axis.
    val_min_1d = float("inf")
    val_max_1d = -float("inf")
    val_min_2d = float("inf")
    val_max_2d = -float("inf")

    source['reds'] = []  # reset the sources for red markers
    plots_to_do = len(active_list)*(len(active_list)+1)/2  # Triangular numbers
    current_plot = 0  # count how many plots we have calculated so far

    # First we calculate xi, yi and zi for each combination of parameters
    for i_active in range(len(active_list)):
        row = []
        for j_active in range(len(active_list)):
            i = active_list[i_active]
            j = active_list[j_active]
            if j > i:
                # We only plot the lower left half of the grid,
                # to avoid duplicates.
                break
            elif i == j:  # Diagonal
                # Passing j = None to dependence
                # makes it calculate a diagonal plot
                xi, yi, _ = dependence(
                    space,
                    model,
                    i,
                    j=None,
                    sample_points=None,
                    n_samples=250,
                    n_points=n_points,
                    x_eval=dependence_eval)
                row.append({"xi": xi, "yi": yi})

                if np.min(yi) < val_min_1d:
                    val_min_1d = np.min(yi)
                if np.max(yi) > val_max_1d:
                    val_max_1d = np.max(yi)

            else:  # Contour plot
                xi, yi, zi = dependence(
                    space,
                    model,
                    i,
                    j=j,
                    sample_points=None,
                    n_samples=50,
                    n_points=n_points,
                    x_eval=dependence_eval)
                row.append({"xi": xi, "yi": yi, "zi": zi})
                if np.min(zi) < val_min_2d:
                    val_min_2d = np.min(zi)
                if np.max(zi) > val_max_2d:
                    val_max_2d = np.max(zi)
            current_plot += 1
            # We put the progress into the div
            # where we normally show the y-values (confidence bounds)
            y_value.text = """<font size="5"><b><br> Calculating objective.
                Please wait. <br> """ + \
                str(int(current_plot))+" / " + \
                str(int(plots_to_do)) + """</font>"""

        plots_data.append(row)

    y_value.text = """<font size="5"><b><br> Drawing.
        Please wait. <br> """ + \
        str(int(current_plot)) + " / " + str(int(plots_to_do)) + """</font>"""
    for i_active in range(len(active_list)):
        # Only plot the selected parameters
        source['reds'].append([])
        plots.append([])
        for j_active in range(len(active_list)):
            i = active_list[i_active]
            j = active_list[j_active]
            if j > i:
                # We only plot the lower left half of the grid,
                # to avoid duplicates.
                break
            elif i == j:  # Diagonal
                # Passing j = None to dependence
                # makes it calculate a diagonal plot
                xi = plots_data[i_active][j_active]["xi"]
                yi = plots_data[i_active][j_active]["yi"]

                if iscat[i]:  # Categorical
                    x_range = space.dimensions[i].categories
                    # Convert integers to catogorical strings
                    xi = [bounds[i][ind] for ind in xi]
                else:  # Numerical
                    x_range = [bounds[i][0], bounds[i][1]]

                # Source red is what we end up appending
                # to the global source variable
                # So the location of the red line can be changed interactively
                source_red = Span(location=red_vals[i],
                                  dimension='height',
                                  line_color='red',
                                  line_width=3)

                plot = figure(plot_height=200,
                              plot_width=250,
                              tools='',
                              x_range=x_range,
                              y_range=[val_min_1d, val_max_1d])
                source_line = ColumnDataSource(data=dict(x=xi, y=yi))
                plot.line('x', 'y', source=source_line,
                          line_width=3, line_alpha=0.6)
                # Add span i.e red line to plot
                plot.add_layout(source_red)
                # update max and minimum y _values

            else:  # Contour plot
                xi = plots_data[i_active][j_active]["xi"]
                yi = plots_data[i_active][j_active]["yi"]
                zi = plots_data[i_active][j_active]["zi"]

                if iscat[j]:  # check if values are categorical
                    # Convert integers to catogorical strings
                    xi = [bounds[j][ind] for ind in xi]
                    # At what coordinate in plot to anchor the contour image.
                    # In case of categorical this should be set to 0.
                    x_anchor = 0
                    # The size the image should take up in the plot
                    x_span = len(result.space.dimensions[j].categories)
                    # Range for axis
                    x_range = space.dimensions[j].categories
                else:
                    x_anchor = bounds[j][0]
                    x_span = bounds[j][1]-bounds[j][0]
                    x_range = [np.min(xi), np.max(xi)]
                if iscat[i]:  # check if values are categorical
                    yi = [bounds[i][ind] for ind in yi]
                    y_range = space.dimensions[i].categories
                    y_anchor = 0
                    y_span = len(result.space.dimensions[i].categories)
                else:
                    y_anchor = bounds[i][0]
                    y_span = bounds[i][1]-bounds[i][0]
                    y_range = [np.min(yi), np.max(yi)]

                plot = figure(plot_height=200, plot_width=250,
                              x_range=x_range, y_range=y_range, tools='')

                if use_same_color_map:
                    cmap = [val_min_2d, val_max_2d]
                else:
                    cmap = None
                # Get an rgba contour image from matplotlib
                # as bokeh does not support contour plots
                im = get_plt_contour_as_rgba(xi,
                                             yi,
                                             zi,
                                             confidence,
                                             cmap=cmap,
                                             draw_confidence=draw_confidence)
                plot.image_rgba(image=[im], x=x_anchor,
                                y=y_anchor, dw=x_span, dh=y_span)
                # x and y samples are the coordinates of the parameter values
                # that have been sampled during the creation of the model.
                x_samples = [val[j] for val in result.x_iters]
                y_samples = [val[i] for val in result.x_iters]
                source_samples = ColumnDataSource(
                    data=dict(x=x_samples, y=y_samples))
                source_red = ColumnDataSource(
                    data=dict(x=[red_vals[j]], y=[red_vals[i]]))
                # We plot samples as black circles and the evaluation marker
                # as a red circle.
                plot.circle(x='x', y='y', source=source_samples,
                            size=2, color="black", alpha=0.5)
                plot.circle(x='x', y='y', source=source_red,
                            size=5, color="red", alpha=1)

            # We rotate the categorical labels slighty
            # so they take up less space
            if iscat[j]:
                plot.xaxis.major_label_orientation = 0.3
            if iscat[i] and i != j:
                # In case of diagonal, the y-labels are numbers
                # and therefore should not be rotated.
                plot.yaxis.major_label_orientation = 1.2

            plot.toolbar.logo = None  # Remove the bokeh logo fom figures
            plots[i_active].append(plot)
            source['reds'][i_active].append(source_red)

    # Setting the same y-range for all diagonal plots for easier comparison
    # for i in range(len(plots)):
    # plots[i][i].y_range = Range1d(y_min,y_max)
    if use_same_color_map:
        colorbar_interval = [val_min_2d, val_max_2d]
    else:
        colorbar_interval = None

    return plots, colorbar_interval


def get_plots_layout(layout,
                     result,
                     active_list,
                     n_points,
                     x_eval,
                     confidence):

    """Description of get_plots_layout

    Create the layout using the lists of plots and selectors.
    The layout consists of rows that we append plots and selectors to. The
    selectors should be on top of each diagonal plot, and therefore the layout
    can be regarded as a (N+1,N) grid to make room for the top selector. The
    top row therefore consists of only on selector and the bottom row has no
    selector as the selector for the last diagonal was added in the second
    last row.

    """

    global x_eval_selectors
    plots, colorbar_interval = get_plot_list(
        layout, result, active_list, n_points, x_eval, confidence)
    x_eval_selectors = get_x_eval_selectors_list(result, active_list, x_eval)

    rows = []
    # The selector in the top left corner
    rows.append(row(x_eval_selectors[0]))
    for i in range(len(plots)):
        if i == len(x_eval_selectors)-1:  # Last row
            rows.append(row(*plots[i]))
        else:
            rows.append(row(*plots[i], x_eval_selectors[i+1]))
    # Create a column with all the rows
    plot_layout = column(*rows)
    return plot_layout, colorbar_interval


def get_x_eval_selectors_list(result, active_list, x_eval):

    """Description of get_x_eval_selectors_list

    Returns a list of selectors. The selectors are sliders for numerical
    values and dropdown menus ("Select" object) for categorical values. The
    selectors are interactive with callbacks everytime a changed by the user.

    """

    global x_eval_selectors_values

    bounds = result.space.bounds
    # Used for defining what values can be selected
    x_eval_selectors = []
    n = 0
    # Index of the plots. Example: If only parameter 3 and 5 is being plotted
    # the selectors for these parameters still have index n = 0 and n= 1.
    for i in active_list:
        # Only get selecters that is going to be shown in GUI
        if isinstance(result.space.dimensions[i], Categorical):  # Categorical
            cats = list(result.space.dimensions[i].categories)  # Categories
            # Create a "Select" object which is a type of dropdown menu
            # This object gets a title equal to the parameter number,
            # and the value is set to x_eval
            select = Select(title='X'+str(i),
                            value=x_eval[i],
                            options=cats,
                            width=200,
                            height=45)
            # Here we define a callback that updates the appropiate
            # red markers by changing with the current value of the selector
            # by changing the global "source" variable.
            # The callback function is written in javascript
            select.js_on_change(
                'value',
                CustomJS(args=dict(source=source, n=n, cats=cats), code="""
                         // Convert categorical to index
                         var ind = cats.indexOf(cb_obj.value);
                         // Change red line in diagonal plots
                         source['reds'][n][n]['location'] = ind + 0.5;
                         // Change red markers in all contour plots
                         // First we change the plots in a vertical direction
                         for (i = n+1; i < source.reds.length; i++) {
                                 source.reds[i][n].data.x = [ind + 0.5] ;
                                 source.reds[i][n].change.emit()
                                 }
                         // Then in a horizontal direction
                         for (j = 0; j < n; j++) {
                                 source.reds[n][j].data.y = [ind + 0.5] ;
                                 source.reds[n][j].change.emit();
                                 }
                         """)
                )
            x_eval_selectors.append(select)
            # We update the global selector values
            x_eval_selectors_values[i] = x_eval[i]
        else:  # Numerical
            # For numerical values we create a slider
            # Minimum and maximum values for slider
            start = bounds[i][0]
            end = bounds[i][1]
            # We change the stepsize according to the range of the slider
            step = get_step_size(start, end)
            slider = Slider(start=start,
                            end=end,
                            value=x_eval[i],
                            step=step,
                            title='X'+str(i),
                            width=200,
                            height=30)
            # javascript callback function that gets called everytime a user
            # changes the slider value.
            slider.js_on_change(
                'value',
                CustomJS(args=dict(source=source, n=n), code="""
                         source.reds[n][n].location = cb_obj.value;
                         source.reds[n][n].change.emit()
                         for (i = n+1; i < source.reds.length; i++) {
                                 source.reds[i][n].data.x = [cb_obj.value] ;
                                 source.reds[i][n].change.emit();
                                 }
                         for (j = 0; j < n; j++) {
                                 source.reds[n][j].data.y = [cb_obj.value] ;
                                 source.reds[n][j].change.emit();
                                 }
                         """)
                )
            x_eval_selectors.append(slider)
            x_eval_selectors_values[i] = x_eval[i]
        n += 1
    return x_eval_selectors


def get_colorbar_as_rgba():
    # plots a colorbar without axes. Code stolen from

    fig, ax = plt.subplots()
    col_map = plt.get_cmap('viridis_r')
    mpl.colorbar.ColorbarBase(ax, cmap=col_map, orientation='vertical')
    fig.subplots_adjust(top=1.01, bottom=0, left=0, right=1)
    plt.axis('off')
    # As for a more fancy example, you can also give an axes by hand:
    fig.canvas.draw()
    # Grab the pixel buffer and dump it into a numpy array
    X = np.array(fig.canvas.renderer._renderer)
    xdim = X.shape[1]
    ydim = X.shape[0]
    # Converting image so that bokeh's image_rgba can read it.
    img = np.empty((ydim, xdim), dtype=np.uint32)
    view = img.view(dtype=np.uint8).reshape((ydim, xdim, 4))
    view[:, :, :] = np.flipud(X)
    plt.close()
    return img


def get_plt_contour_as_rgba(xi,
                            yi,
                            zi,
                            confidence,
                            cmap=None,
                            draw_confidence=None):

    """Description of get_plt_contour_as_rgba

    Returns a matplotlib contour plot as an rgba image. We create a matplotlib
    figure and draws it so we can capture the figure as an image.

    """

    fig = plt.figure()
    ax = fig.add_axes([0., 0., 1., 1.])
    ax = plt.gca()
    if cmap:
        ax.contourf(xi, yi, zi, 10, locator=None,
                    cmap='viridis_r', vmin=cmap[0], vmax=cmap[1])
    else:
        ax.contourf(xi, yi, zi, 10, locator=None, cmap='viridis_r')
    if draw_confidence:
        ax.contour(xi,
                   yi,
                   zi,
                   levels=[confidence[1]],
                   locator=None,
                   colors='black',
                   linestyles=('--',),
                   linewidths=(3,))
    plt.axis('off')
    fig.canvas.draw()
    # Grab the pixel buffer and dump it into a numpy array
    X = np.array(fig.canvas.renderer._renderer)
    xdim = X.shape[1]
    ydim = X.shape[0]
    # Converting image so that bokeh's image_rgba can read it.
    img = np.empty((ydim, xdim), dtype=np.uint32)
    view = img.view(dtype=np.uint8).reshape((ydim, xdim, 4))
    view[:, :, :] = np.flipud(X)
    plt.close()
    return img


def get_x_eval(result, active_list):
    # Returns the evaluation values that are defined by
    # the evaluation method dropdown menu
    _, _, iscat = _map_categories(result.space, [result.x], result.x)
    eval_method = dropdown_eval_method.value  # Get evaluation method from GUI

    if eval_method == 'Exp min' and any(iscat):
        # Expected_minimum does not support categorical values
        eval_method = "Exp min rand"
    if eval_method == 'Result':
        x = result.x
    elif eval_method == 'Sliders':
        x = get_x_eval_selectors_values()
    elif eval_method == 'Exp min' and not any(iscat):
        x, _ = expected_minimum(result, n_random_starts=10, random_state=None)
    elif eval_method == 'Exp min rand':
        x = expected_minimum_random_sampling(result,
                                             n_random_starts=100000,
                                             random_state=None)
    else:
        ValueError('Could not find evalmethod from dropdown menu')
    return x


def get_step_size(start, end):
    # Returns the stepsize to be used for sliders
    # the stepsize will always be of the form 10**x
    range_log = round((math.log(end-start, 10)))
    # The bigger the range the bigger the stepsize
    step = 10**(range_log-3)
    return step


def modify_doc(doc):
    # Add layout to document
    doc.add_root(layout)
# Update once to initialize message
    handle_button_generate(layout, result)


def start(parsed_result):
    # Set the global variables using the parsed "result"
    set_globals(parsed_result)
    # Start server
    server = Server({'/': modify_doc}, num_procs=1)
    server.start()
    print('Opening Bokeh application on http://localhost:5006/')
    server.io_loop.add_callback(server.show, "/")
    server.io_loop.start()



"""Forest based minimization algorithms."""

from .base import base_minimize


def forest_minimize(func, dimensions, base_estimator="ET", n_calls=100,
                    n_random_starts=10, acq_func="EI",
                    x0=None, y0=None, random_state=None, verbose=False,
                    callback=None, n_points=10000, xi=0.01, kappa=1.96,
                    n_jobs=1):
    """Sequential optimisation using decision trees.

    A tree based regression model is used to model the expensive to evaluate
    function `func`. The model is improved by sequentially evaluating
    the expensive function at the next best point. Thereby finding the
    minimum of `func` with as few evaluations as possible.

    The total number of evaluations, `n_calls`, are performed like the
    following. If `x0` is provided but not `y0`, then the elements of `x0`
    are first evaluated, followed by `n_random_starts` evaluations.
    Finally, `n_calls - len(x0) - n_random_starts` evaluations are
    made guided by the surrogate model. If `x0` and `y0` are both
    provided then `n_random_starts` evaluations are first made then
    `n_calls - n_random_starts` subsequent evaluations are made
    guided by the surrogate model.

    Parameters
    ----------
    * `func` [callable]:
        Function to minimize. Should take a single list of parameters
        and return the objective value.

        If you have a search-space where all dimensions have names,
        then you can use `ProcessOptimizer.utils.use_named_args` as a decorator
        on your objective function, in order to call it directly
        with the named arguments. See `use_named_args` for an example.

    * `dimensions` [list, shape=(n_dims,)]:
        List of search space dimensions.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, prior)` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

         NOTE: The upper and lower bounds are inclusive for `Integer`
         dimensions.

    * `base_estimator` [string or `Regressor`, default=`"ET"`]:
        The regressor to use as surrogate model. Can be either

        - `"RF"` for random forest regressor
        - `"ET"` for extra trees regressor
        - instance of regressor with support for `return_std` in its predict
          method

        The predefined models are initilized with good defaults. If you
        want to adjust the model parameters pass your own instance of
        a regressor which returns the mean and standard deviation when
        making predictions.

    * `n_calls` [int, default=100]:
        Number of calls to `func`.

    * `n_random_starts` [int, default=10]:
        Number of evaluations of `func` with random points before
        approximating it with `base_estimator`.

    * `acq_func` [string, default=`"LCB"`]:
        Function to minimize over the forest posterior. Can be either

        - `"LCB"` for lower confidence bound.
        - `"EI"` for negative expected improvement.
        - `"PI"` for negative probability of improvement.
        - `"EIps" for negated expected improvement per second to take into
          account the function compute time. Then, the objective function is
          assumed to return two values, the first being the objective value and
          the second being the time taken in seconds.
        - `"PIps"` for negated probability of improvement per second. The
          return type of the objective function is assumed to be similar to
          that of `"EIps"`

    * `x0` [list, list of lists or `None`]:
        Initial input points.

        - If it is a list of lists, use it as a list of input points.
        - If it is a list, use it as a single initial input point.
        - If it is `None`, no initial input points are used.

    * `y0` [list, scalar or `None`]:
        Evaluation of initial input points.

        - If it is a list, then it corresponds to evaluations of the function
          at each element of `x0` : the i-th element of `y0` corresponds
          to the function evaluated at the i-th element of `x0`.
        - If it is a scalar, then it corresponds to the evaluation of the
          function at `x0`.
        - If it is None and `x0` is provided, then the function is evaluated
          at each element of `x0`.

    * `random_state` [int, RandomState instance, or None (default)]:
        Set random state to something other than None for reproducible
        results.

    * `verbose` [boolean, default=False]:
        Control the verbosity. It is advised to set the verbosity to True
        for long optimization runs.

    * `callback` [callable, optional]
        If provided, then `callback(res)` is called after call to func.

    * `n_points` [int, default=10000]:
        Number of points to sample when minimizing the acquisition function.

    * `xi` [float, default=0.01]:
        Controls how much improvement one wants over the previous best
        values. Used when the acquisition is either `"EI"` or `"PI"`.

    * `kappa` [float, default=1.96]:
        Controls how much of the variance in the predicted values should be
        taken into account. If set to be very high, then we are favouring
        exploration over exploitation and vice versa.
        Used when the acquisition is `"LCB"`.

    * `n_jobs` [int, default=1]:
        The number of jobs to run in parallel for `fit` and `predict`.
        If -1, then the number of jobs is set to the number of cores.

    Returns
    -------
    * `res` [`OptimizeResult`, scipy object]:
        The optimization result returned as a OptimizeResult object.
        Important attributes are:

        - `x` [list]: location of the minimum.
        - `fun` [float]: function value at the minimum.
        - `models`: surrogate models used for each iteration.
        - `x_iters` [list of lists]: location of function evaluation for each
           iteration.
        - `func_vals` [array]: function value for each iteration.
        - `space` [Space]: the optimization space.
        - `specs` [dict]`: the call specifications.

        For more details related to the OptimizeResult object, refer
        http://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html
    """
    return base_minimize(func, dimensions, base_estimator,
                         n_calls=n_calls, n_points=n_points,
                         n_random_starts=n_random_starts,
                         x0=x0, y0=y0, random_state=random_state,
                         acq_func=acq_func,
                         xi=xi, kappa=kappa, verbose=verbose,
                         callback=callback, acq_optimizer="sampling")
